# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import uuid

import numpy as np
import scipp as sc
from matplotlib.dates import date2num
from matplotlib.lines import Line2D
from numpy.typing import ArrayLike

from ..common import make_line_data
from .canvas import Canvas
from .utils import make_legend, parse_dicts_in_kwargs


def _to_float(x):
    return date2num(x) if np.issubdtype(x.dtype, np.datetime64) else x


class Line:
    """
    Artist to represent one-dimensional data.
    If the coordinate is bin centers, the line is (by default) a set of markers.
    If the coordinate is bin edges, the line is a step function.

    Parameters
    ----------
    canvas:
        The canvas that will display the line.
    data:
        The initial data to create the line from.
    number:
        The canvas keeps track of how many lines have been added to it. This number is
        used to set the color and marker parameters of the line.
    """

    def __init__(self, canvas: Canvas, data: sc.DataArray, number: int = 0, **kwargs):
        self._canvas = canvas
        self._ax = self._canvas.ax
        self._data = data
        # Because all keyword arguments from the figure are forwarded to both the canvas
        # and the line, we need to remove the arguments that belong to the canvas.
        kwargs.pop('ax', None)

        args = parse_dicts_in_kwargs(kwargs, name=data.name)

        self._line = None
        self._mask = None
        self._error = None
        self._unit = None
        self.label = data.name
        self._dim = self._data.dim
        self._unit = self._data.unit
        self._coord = self._data.coords[self._dim]
        self._id = uuid.uuid4().hex

        aliases = {'ls': 'linestyle', 'lw': 'linewidth', 'c': 'color'}
        for key, alias in aliases.items():
            if key in args:
                args[alias] = args.pop(key)

        self._make_line(
            data=make_line_data(data=self._data, dim=self._dim),
            number=number,
            **args,
        )

    def _make_line(
        self,
        data: dict,
        number: int,
        errorbars: bool = True,
        mask_color: str = 'black',
        **kwargs,
    ):
        """
        Create either plot markers or a step function, depending on whether the data
        contains bin edges or not.

        Parameters
        ----------
        data:
            A dictionary containing data entries that have been pre-processed to be in
            a format that Matplotlib can directly use.
        number:
            The line number to set colors and marker style.
        errorbars:
            Show errorbars if ``True``.
        mask_color:
            The color to be used to represent the masks.
        **kwargs:
            The kwargs are forwarded to:

            - ``matplotlib.pyplot.plot`` for data with a non bin-edge coordinate
            - ``matplotlib.pyplot.step`` for data with a bin-edge coordinate
        """

        default_step_style = {
            'linestyle': 'solid',
            'linewidth': 1.5,
            'color': f'C{number}',
        }
        markers = list(Line2D.markers.keys())
        default_plot_style = {
            'linestyle': 'none',
            'linewidth': 1.5,
            'marker': markers[(number + 2) % len(markers)],
            'color': f'C{number}',
        }

        if data["hist"]:
            self._line = self._ax.step(
                data['values']['x'],
                data['values']['y'],
                label=self.label,
                zorder=10,
                **{**default_step_style, **kwargs},
            )[0]

            self._mask = self._ax.step(data['mask']['x'], data['mask']['y'])[0]
            self._mask.update_from(self._line)
            self._mask.set_color(mask_color)
            self._mask.set_label(None)
            self._mask.set_linewidth(self._mask.get_linewidth() * 3)
            self._mask.set_zorder(self._mask.get_zorder() - 1)
            self._mask.set_visible(data['mask']['visible'])
        else:
            self._line = self._ax.plot(
                data['values']['x'],
                data['values']['y'],
                label=self.label,
                zorder=10,
                **{**default_plot_style, **kwargs},
            )[0]
            self._mask = self._ax.plot(
                data['mask']['x'],
                data['mask']['y'],
                zorder=11,
                mec=mask_color,
                mfc="None",
                mew=3.0,
                linestyle="none",
                marker=self._line.get_marker(),
                visible=data['mask']['visible'],
            )[0]

        line_mask = ~np.isnan(data['mask']['y'])
        self._line._plopp_mask = line_mask

        # Add error bars
        if errorbars and (data['stddevs'] is not None):
            self._error = self._ax.errorbar(
                data['stddevs']['x'],
                data['stddevs']['y'],
                yerr=data['stddevs']['e'],
                color=self._line.get_color(),
                zorder=10,
                fmt="none",
            )
            # Set the selection mask on the line collection that makes the segments
            self._error[2][0]._plopp_mask = line_mask[1:] if data["hist"] else line_mask

        if self.label and self._canvas._legend:
            self._ax.legend(**make_legend(self._canvas._legend))

    def update(self, new_values: sc.DataArray):
        """
        Update the x and y positions of the data points from new data.

        Parameters
        ----------
        new_values:
            New data to update the line values, masks, errorbars from.
        """
        self._data = new_values
        new_values = make_line_data(data=self._data, dim=self._dim)

        self._line.set_data(new_values['values']['x'], new_values['values']['y'])
        self._mask.set_data(new_values['mask']['x'], new_values['mask']['y'])
        self._mask.set_visible(new_values['mask']['visible'])

        line_mask = ~np.isnan(new_values['mask']['y'])
        self._line._plopp_mask = line_mask

        if (self._error is not None) and (new_values['stddevs'] is not None):
            coll = self._error.get_children()[0]
            coll.set_segments(
                self._change_segments_y(
                    _to_float(new_values['stddevs']['x']),
                    _to_float(new_values['stddevs']['y']),
                    _to_float(new_values['stddevs']['e']),
                )
            )
            self._error[2][0]._plopp_mask = (
                line_mask[1:] if new_values["hist"] else line_mask
            )

    def _change_segments_y(self, x: ArrayLike, y: ArrayLike, e: ArrayLike) -> ArrayLike:
        """
        Update the positions of the errorbars when `update_data` is called.
        """
        arr1 = np.repeat(x, 2)
        arr2 = np.array([y - e, y + e]).T.flatten()
        return np.array([arr1, arr2]).T.flatten().reshape(len(y), 2, 2)

    def remove(self):
        """
        Remove the line, masks and errorbar artists from the canvas.
        """
        self._line.remove()
        self._mask.remove()
        if self._error is not None:
            self._error.remove()

    @property
    def color(self):
        """
        The line color.
        """
        return self._line.get_color()

    @color.setter
    def color(self, val):
        self._line.set_color(val)
        if self._error is not None:
            for artist in self._error.get_children():
                artist.set_color(val)
        self._canvas.draw()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)


import scipp as sc
from numpy import ndarray

from ..core import Node
from ..core.typing import FigureLike
from ..graphics import linefigure
from .common import to_variable


def _make_data_array(x: sc.Variable, y: sc.Variable) -> sc.DataArray:
    """
    Make a data array from the supplied variables, using ``x`` as the coordinate and
    ``y`` as the data.

    Parameters
    ----------
    x:
        The variable to use as the coordinate.
    y:
        The variable to use as the data.
    """
    return sc.DataArray(data=y, coords={x.dim: x})


def xyplot(
    x: sc.Variable | ndarray | list | Node,
    y: sc.Variable | ndarray | list | Node,
    **kwargs,
) -> FigureLike:
    """
    Make a one-dimensional plot of one variable ``y`` as a function of another ``x``.

    .. versionadded:: 23.10.0

    Parameters
    ----------
    x:
        The variable to use as the coordinates for the horizontal axis.
        Must be one-dimensional.
    y:
        The variable to use as the data for the vertical axis. Must be one-dimensional.
    **kwargs:
        See :py:func:`plopp.plot`.
    """
    x = Node(to_variable, x)
    y = Node(to_variable, y)
    dim = x().dim
    if dim != y().dim:
        raise sc.DimensionError("Dimensions of x and y must match")
    return linefigure(Node(_make_data_array, x=x, y=y), **kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import os
import tempfile

import numpy as np
import pytest
import scipp as sc

import plopp as pp
from plopp.data.testing import data_array, dataset


def test_plot_ndarray():
    pp.plot(np.arange(50.0))


def test_plot_ndarray_int():
    pp.plot(np.arange(50))


def test_plot_list():
    pp.plot([1.0, 2.0, 3.0, 4.0, 5.0])


def test_plot_list_int():
    pp.plot([1, 2, 3, 4, 5])


def test_plot_variable():
    pp.plot(sc.arange('x', 50.0))


def test_plot_data_array():
    pp.plot(data_array(ndim=1))


def test_plot_data_array_missing_coords():
    da = sc.data.table_xyz(100)
    pp.plot(da)
    assert 'row' not in da.coords


def test_plot_dataset():
    ds = dataset(ndim=1)
    pp.plot(ds)


def test_plot_dict_of_ndarrays():
    pp.plot({'a': np.arange(50.0), 'b': np.arange(60.0)})


def test_plot_dict_of_lists():
    pp.plot({'a': [1, 2, 3, 4], 'b': [4, 5, 6, 7]})


def test_plot_dict_of_variables():
    pp.plot({'a': sc.arange('x', 50.0), 'b': sc.arange('x', 60.0)})


def test_plot_dict_of_data_arrays():
    ds = dataset(ndim=1)
    pp.plot({'a': ds['a'], 'b': ds['b']})


def test_plot_from_node():
    da = data_array(ndim=1)
    pp.plot(pp.Node(da))


def test_plot_multiple_inputs_raises():
    da = data_array(ndim=1)
    with pytest.raises(TypeError, match='takes 1 positional argument'):
        pp.plot(da, 3.3 * da)


def test_plot_dict_of_nodes():
    a = data_array(ndim=1)
    b = 6.7 * a
    pp.plot({'a': pp.Node(a), 'b': pp.Node(b)})


def test_plot_mixing_raw_data_and_nodes():
    a = data_array(ndim=1)
    b = 13.3 * a
    pp.plot({'a': a, 'b': pp.Node(b)})
    pp.plot({'a': pp.Node(a), 'b': b})


def test_plot_coord_with_no_unit():
    da = sc.data.table_xyz(100).bin(x=10).bins.mean()
    da.coords['x'].unit = None
    pp.plot(da)


def test_plot_rejects_large_input():
    error_match = "may take very long or use an excessive amount of memory"
    with pytest.raises(ValueError, match=error_match):
        pp.plot(np.random.random(1_100_000))
    with pytest.raises(ValueError, match=error_match):
        pp.plot(np.random.random((3000, 2500)))


def test_plot_ignore_size_disables_size_check():
    pp.plot(np.random.random(1_100_000), ignore_size=True)


def test_plot_with_non_dimensional_unsorted_coord_does_not_warn():
    da = data_array(ndim=1)
    da.coords['aux'] = sc.sin(sc.arange(da.dim, 50.0, unit='rad'))
    pp.plot(da)


def _get_artist(fig, n=0):
    artists = fig.artists
    keys = list(artists.keys())
    return artists[keys[n]]


def _get_line(fig, n=0):
    return _get_artist(fig=fig, n=n)._line


def test_kwarg_linecolor():
    da = data_array(ndim=1)
    p = pp.plot(da, color='r')
    assert _get_line(p).get_color() == 'r'
    p = pp.plot(da, c='b')
    assert _get_line(p).get_color() == 'b'


def test_kwarg_linestyle():
    da = data_array(ndim=1)
    p = pp.plot(da, linestyle='solid')
    assert _get_line(p).get_linestyle() == '-'
    p = pp.plot(da, ls='dashed')
    assert _get_line(p).get_linestyle() == '--'


def test_kwarg_linewidth():
    da = data_array(ndim=1)
    p = pp.plot(da, linewidth=3)
    assert _get_line(p).get_linewidth() == 3
    p = pp.plot(da, lw=5)
    assert _get_line(p).get_linewidth() == 5


def test_kwarg_marker():
    da = data_array(ndim=1)
    p = pp.plot(da, marker='+')
    assert _get_line(p).get_marker() == '+'


def test_kwarg_norm():
    da = data_array(ndim=1)
    p = pp.plot(da, norm='log')
    assert p.canvas.ax.get_yscale() == 'log'


def test_kwarg_scale():
    da = data_array(ndim=1)
    p = pp.plot(da, scale={'xx': 'log'})
    assert p.canvas.ax.get_xscale() == 'log'
    assert p.canvas.ax.get_yscale() == 'linear'


def test_kwarg_for_two_lines():
    ds = dataset(ndim=1)
    p = pp.plot(ds, color='r')
    assert _get_line(p, 0).get_color() == 'r'
    assert _get_line(p, 1).get_color() == 'r'


def test_kwarg_as_dict():
    ds = dataset(ndim=1)
    p = pp.plot(ds, color={'a': 'red', 'b': 'black'})
    assert _get_line(p, 0).get_color() == 'red'
    assert _get_line(p, 1).get_color() == 'black'
    da = data_array(ndim=1)
    p = pp.plot({'a': da, 'b': 0.2 * da}, color={'a': 'red', 'b': 'black'})
    assert _get_line(p, 0).get_color() == 'red'
    assert _get_line(p, 1).get_color() == 'black'


def test_raises_ValueError_when_given_binned_data():
    da = sc.data.table_xyz(100).bin(x=10)
    with pytest.raises(ValueError, match='Cannot plot binned data'):
        pp.plot(da)


def test_raises_ValueError_when_given_unsupported_data_type():
    a = data_array(ndim=1)
    b = a * 2.0
    c = a * 3.0
    d = a * 4.0
    nested_dict = {'group1': {'a': a, 'b': b}, 'group2': {'c': c, 'd': d}}
    with pytest.raises(TypeError, match='Cannot convert input of type'):
        pp.plot(nested_dict)


def test_use_non_dimension_coords_dataset():
    ds = dataset(ndim=1)
    ds.coords['xx2'] = 6.6 * ds.coords['xx']
    p = pp.plot(ds, coords=['xx2'])
    assert p.canvas.dims['x'] == 'xx2'
    assert p.canvas.xmax > 6.6 * ds.coords['xx'].max().value


@pytest.mark.parametrize('ext', ['jpg', 'png', 'pdf', 'svg'])
def test_save_to_disk_1d(ext):
    da = data_array(ndim=1)
    fig = pp.plot(da)
    with tempfile.TemporaryDirectory() as path:
        fname = os.path.join(path, f'plopp_fig1d.{ext}')
        fig.save(filename=fname)
        assert os.path.isfile(fname)


def test_save_to_disk_with_bad_extension_raises():
    da = data_array(ndim=1)
    fig = pp.plot(da)
    with pytest.raises(ValueError, match='txt'):
        fig.save(filename='plopp_fig1d.txt')


def test_plot_xarray_data_array_1d():
    import xarray as xr

    N = 50
    data = np.random.random(N)
    time = np.arange(float(N))
    da = xr.DataArray(data, coords={'time': time}, dims=['time'])
    p = pp.plot(da)
    assert p.canvas.dims['x'] == 'time'
    assert p.canvas.units['x'] == 'dimensionless'
    assert p.canvas.units['y'] == 'dimensionless'


def test_plot_xarray_dataset():
    import xarray as xr

    N = 50
    temp = 15 + 8 * np.random.random(N)
    precip = 10 * np.random.random(N)
    ds = xr.Dataset(
        {
            "temperature": (["time"], temp),
            "precipitation": (["time"], precip),
        },
        coords={"time": np.arange(50)},
    )
    p = pp.plot(ds)
    assert p.canvas.dims['x'] == 'time'
    assert p.canvas.units['x'] == 'dimensionless'
    assert p.canvas.units['y'] == 'dimensionless'
    assert len(p._view.artists) == 2


def test_plot_pandas_series():
    import pandas as pd

    s = pd.Series(np.arange(100.0), name='MyDataSeries')
    p = pp.plot(s)
    assert p.canvas.dims['x'] == 'row'
    assert next(iter(p._view.artists.values())).label == 'MyDataSeries'


def test_plot_pandas_dataframe():
    import pandas as pd

    df = pd.DataFrame(
        {
            'A': np.arange(50.0),
            'B': 1.5 * np.arange(50),
            'C': np.random.random(50),
            'D': np.random.normal(size=50),
        }
    )
    p = pp.plot(df)
    assert p.canvas.dims['x'] == 'row'
    assert len(p._view.artists) == 4


def test_hide_legend():
    da1 = data_array(ndim=1)
    da2 = da1 * 3.3
    p = pp.plot({'a': da1, 'b': da2}, legend=False)
    leg = p.ax.get_legend()
    assert leg is None


def test_legend_location():
    da1 = data_array(ndim=1)
    da2 = da1 * 3.3
    data = {'a': da1, 'b': da2}
    leg1 = pp.plot(data, legend=(0.5, 0.5)).ax.get_legend().get_window_extent().bounds
    leg2 = pp.plot(data, legend=(0.9, 0.5)).ax.get_legend().get_window_extent().bounds
    leg3 = pp.plot(data, legend=(0.5, 0.9)).ax.get_legend().get_window_extent().bounds
    assert leg2[0] > leg1[0]
    assert leg2[1] == leg1[1]
    assert leg3[1] > leg1[1]
    assert leg3[0] == leg1[0]


def test_hide_legend_bad_type():
    da1 = data_array(ndim=1)
    da2 = da1 * 3.3
    with pytest.raises(TypeError, match='Legend must be a bool, tuple, or a list'):
        pp.plot({'a': da1, 'b': da2}, legend='False')


@pytest.mark.parametrize('Constructor', [dict, sc.Dataset, sc.DataGroup])
def test_names_are_overridden_when_plotting_dicts(Constructor):
    da1 = data_array(ndim=1)
    da2 = da1 * 2
    da1.name = "DA1"
    da2.name = "DA2"
    p = pp.plot(Constructor({'a': da1, 'b': da2}))
    assert p.ax.get_legend().texts[0].get_text() == 'a'
    assert p.ax.get_legend().texts[1].get_text() == 'b'
    artists = list(p._view.artists.values())
    assert artists[0].label == 'a'
    assert artists[1].label == 'b'
    assert da1.name == 'DA1'
    assert da2.name == 'DA2'


def test_plot_1d_ignores_masked_data_for_vertical_range():
    da = data_array(ndim=1)
    da.values[10] = 100
    da.masks['m'] = da.data > sc.scalar(5.0, unit='m/s')
    p = pp.plot(da)
    assert p.canvas.ymax < 100


def test_plot_1d_includes_masked_data_in_horizontal_range():
    da = data_array(ndim=1)
    da.masks['m'] = da.coords['xx'] > sc.scalar(5.0, unit='m')
    p = pp.plot(da)
    # If we check only for > 5, padding may invalidate the test
    assert p.canvas.xmax > 10.0


def test_plot_1d_datetime_coord():
    t = np.arange(
        np.datetime64('2017-03-16T20:58:17'), np.datetime64('2017-03-16T21:15:17'), 20
    )
    time = sc.array(dims=['time'], values=t)
    v = np.random.rand(time.sizes['time'])
    da = sc.DataArray(
        data=sc.array(dims=['time'], values=10 * v, variances=v),
        coords={'time': time},
    )
    pp.plot(da)


def test_plot_1d_datetime_coord_binedges():
    t = np.arange(
        np.datetime64('2017-03-16T20:58:17'), np.datetime64('2017-03-16T21:15:17'), 20
    )
    time = sc.array(dims=['time'], values=t)
    v = np.random.rand(time.sizes['time'] - 1)
    da = sc.DataArray(
        data=sc.array(dims=['time'], values=10 * v, variances=v),
        coords={'time': time},
    )
    pp.plot(da)


def test_plot_1d_datetime_coord_with_mask():
    t = np.arange(
        np.datetime64('2017-03-16T20:58:17'), np.datetime64('2017-03-16T21:15:17'), 20
    )
    time = sc.array(dims=['time'], values=t)
    v = np.random.rand(time.sizes['time'])
    da = sc.DataArray(
        data=sc.array(dims=['time'], values=10 * v, variances=v),
        coords={'time': time},
        masks={'m': time > sc.datetime('2017-03-16T21:10:00')},
    )
    pp.plot(da)


def test_plot_1d_datetime_coord_with_mask_and_binedges():
    t = np.arange(
        np.datetime64('2017-03-16T20:58:17'), np.datetime64('2017-03-16T21:15:17'), 20
    )
    time = sc.array(dims=['time'], values=t)
    v = np.random.rand(time.sizes['time'] - 1)
    da = sc.DataArray(
        data=sc.array(dims=['time'], values=10 * v, variances=v),
        coords={'time': time},
        masks={'m': sc.midpoints(time) > sc.datetime('2017-03-16T21:10:00')},
    )
    pp.plot(da)


def test_plot_1d_datetime_coord_log():
    t = np.arange(
        np.datetime64('2017-03-16T20:58:17'), np.datetime64('2017-03-16T21:15:17'), 20
    )
    time = sc.array(dims=['time'], values=t)
    v = np.random.rand(time.sizes['time'])
    da = sc.DataArray(
        data=sc.array(dims=['time'], values=10 * v, variances=v),
        coords={'time': time},
    )
    pp.plot(da, scale={'time': 'log'})


def test_plot_1d_datetime_coord_log_binedges():
    t = np.arange(
        np.datetime64('2017-03-16T20:58:17'), np.datetime64('2017-03-16T21:15:17'), 20
    )
    time = sc.array(dims=['time'], values=t)
    v = np.random.rand(time.sizes['time'] - 1)
    da = sc.DataArray(
        data=sc.array(dims=['time'], values=10 * v, variances=v),
        coords={'time': time},
    )
    pp.plot(da, scale={'time': 'log'})


def test_plot_1d_datetime_coord_log_with_mask():
    t = np.arange(
        np.datetime64('2017-03-16T20:58:17'), np.datetime64('2017-03-16T21:15:17'), 20
    )
    time = sc.array(dims=['time'], values=t)
    v = np.random.rand(time.sizes['time'])
    da = sc.DataArray(
        data=sc.array(dims=['time'], values=10 * v, variances=v),
        coords={'time': time},
        masks={'m': time > sc.datetime('2017-03-16T21:10:00')},
    )
    pp.plot(da, scale={'time': 'log'})


def test_plot_1d_data_with_errorbars():
    da = data_array(ndim=1, variances=True)
    p = da.plot()
    assert p.canvas.ymin < -1.0
    assert p.canvas.ymax > 1.0


def test_plot_1d_data_with_variances_and_nan_values():
    da = data_array(ndim=1, variances=True)
    da.values[-10:] = np.nan
    p = da.plot()
    assert p.canvas.ymin < -1.0
    assert p.canvas.ymax > 1.0



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import ipywidgets as ipw
import pytest

from plopp.widgets import HBar, VBar


@pytest.mark.parametrize('Bar', [VBar, HBar])
def test_bar_creation(Bar):
    b1 = ipw.Button(description='Button 1')
    b2 = ipw.Button(description='Button 2')
    bar = Bar([b1, b2])
    assert len(bar.children) == 2
    assert bar.children[0] is b1
    assert bar.children[1] is b2


@pytest.mark.parametrize('Bar', [VBar, HBar])
def test_bar_add(Bar):
    b1 = ipw.Button(description='Button 1')
    b2 = ipw.Button(description='Button 2')
    bar = Bar([b1])
    assert len(bar.children) == 1
    bar.add(b2)
    assert len(bar.children) == 2
    assert bar.children[0] is b1
    assert bar.children[1] is b2


@pytest.mark.parametrize('Bar', [VBar, HBar])
def test_bar_remove(Bar):
    b1 = ipw.Button(description='Button 1')
    b2 = ipw.Button(description='Button 2')
    bar = Bar([b1, b2])
    assert len(bar.children) == 2
    assert bar.children[0] is b1
    assert bar.children[1] is b2
    bar.remove(b1)
    assert len(bar.children) == 1
    assert bar.children[0] is b2
    bar = Bar([b1, b2])
    bar.remove(b2)
    assert len(bar.children) == 1
    assert bar.children[0] is b1


@pytest.mark.parametrize('Bar', [VBar, HBar])
def test_bar_indexing(Bar):
    b1 = ipw.Button(description='Button 1')
    b2 = ipw.Button(description='Button 2')
    bar = Bar([b1, b2])
    assert bar[0] is b1
    assert bar[1] is b2


@pytest.mark.parametrize('Bar', [VBar, HBar])
def test_bar_indexing_range(Bar):
    b1 = ipw.Button(description='Button 1')
    b2 = ipw.Button(description='Button 2')
    b3 = ipw.Button(description='Button 3')
    bar = Bar([b1, b2, b3])
    result = bar[1:3]
    assert isinstance(result, Bar)
    assert len(result.children) == 2
    assert result.children[0] is b2
    assert result.children[1] is b3


@pytest.mark.parametrize('Bar', [VBar, HBar])
def test_bar_indexing_range_step(Bar):
    b1 = ipw.Button(description='Button 1')
    b2 = ipw.Button(description='Button 2')
    b3 = ipw.Button(description='Button 3')
    bar = Bar([b1, b2, b3])
    result = bar[0:3:2]
    assert isinstance(result, Bar)
    assert len(result.children) == 2
    assert result.children[0] is b1
    assert result.children[1] is b3


@pytest.mark.parametrize('Bar', [VBar, HBar])
def test_bar_length(Bar):
    b1 = ipw.Button(description='Button 1')
    b2 = ipw.Button(description='Button 2')
    b3 = ipw.Button(description='Button 3')
    bar = Bar([b1, b2, b3])
    assert len(bar) == 3



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import matplotlib.pyplot as plt
import numpy as np
import scipp as sc

from plopp import Node
from plopp.data.testing import data_array
from plopp.graphics.lineview import LineView


def test_with_string_coord():
    strings = ['a', 'b', 'c', 'd', 'e']
    da = sc.DataArray(
        data=sc.arange('x', 5.0),
        coords={'x': sc.array(dims=['x'], values=strings, unit='m')},
    )
    fig = LineView(Node(da))
    assert [t.get_text() for t in fig.canvas.ax.get_xticklabels()] == strings


def test_with_strings_as_bin_edges():
    strings = ['a', 'b', 'c', 'd', 'e', 'f']
    da = sc.DataArray(
        data=sc.arange('x', 5.0),
        coords={'x': sc.array(dims=['x'], values=strings, unit='m')},
    )
    fig = LineView(Node(da))
    assert [t.get_text() for t in fig.canvas.ax.get_xticklabels()] == strings


def test_figsize():
    da = data_array(ndim=1)
    size = (6.1, 3.3)
    fig = LineView(Node(da), figsize=size)
    assert np.allclose(fig.canvas.fig.get_size_inches(), size)


def test_grid():
    da = data_array(ndim=1)
    fig = LineView(Node(da), grid=True)
    assert fig.canvas.ax.xaxis.get_gridlines()[0].get_visible()


def test_ax():
    _, ax = plt.subplots()
    assert len(ax.lines) == 0
    da = data_array(ndim=1)
    _ = LineView(Node(da), ax=ax)
    assert len(ax.lines) > 0



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from functools import partial
from typing import Any

import ipywidgets as ipw
from matplotlib.colors import to_hex

from ..core import Node, View
from ..core.utils import coord_element_to_string
from .box import VBar


class LineSaveTool(VBar):
    """
    Create a tool that is used to copy and save the currently displayed 1D line on a
    plot.

    Parameters
    ----------
    data_node:
        The node that generates the input data.
    slider_node:
        The node that generates the sliced data.
    fig:
        The figure to which the lines are to be added.
    """

    def __init__(self, data_node: Node, slider_node: Node, fig: View):
        self._data_node = data_node
        self._slider_node = slider_node
        self._fig = fig
        self._lines = {}
        self.button = ipw.Button(description='Save line')
        self.button.on_click(self.save_line)
        self.container = VBar()
        super().__init__([self.button, self.container], layout={'width': '350px'})

    def _update_container(self):
        self.container.children = [line['tool'] for line in self._lines.values()]

    def save_line(self, change: dict[str, Any] | None = None):
        from ..widgets import ColorTool

        data = self._data_node.request_data()
        node = Node(data)
        node.name = f'Save node {len(self._lines)}'
        line_id = node._id
        node.add_view(self._fig._view)
        self._fig._view.render()
        text = ', '.join(
            f'{k}: {coord_element_to_string(data.coords[k])}'
            for k in self._slider_node.request_data()
        )
        line = self._fig.artists[line_id]
        tool = ColorTool(text=text, color=to_hex(line.color))
        self._lines[line_id] = {'line': line, 'tool': tool, 'node': node}
        self._update_container()
        tool.color.observe(
            partial(self.change_line_color, line_id=line_id), names='value'
        )
        tool.button.on_click(partial(self.remove_line, line_id=line_id))

    def change_line_color(self, change: dict[str, Any], line_id: str):
        self._lines[line_id]['line'].color = change['new']

    def remove_line(self, change: dict[str, Any], line_id: str):
        self._lines[line_id]['line'].remove()
        self._lines[line_id]['node'].remove()
        del self._lines[line_id]
        self._update_container()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from collections.abc import Callable
from functools import partial

import ipywidgets as ipw

from .style import BUTTON_LAYOUT


class ButtonTool(ipw.Button):
    """
    Create a button with a callback that is called when the button is clicked.

    Parameters
    ----------
    callback:
        The function that will be called when the button is clicked.
    **kwargs:
        All other kwargs are forwarded to ipywidgets.Button.
    """

    def __init__(self, callback: Callable | None = None, **kwargs):
        super().__init__(**{**BUTTON_LAYOUT, **kwargs})
        self.callback = callback
        self.on_click(self)

    def __call__(self, *ignored):
        self.callback()


class ToggleTool(ipw.ToggleButton):
    """
    Create a toggle button with a callback that is called when the button is
    toggled.

    Parameters
    ----------
    callback:
        The function that will be called when the button is toggled.
    **kwargs:
        All other kwargs are forwarded to ipywidgets.ToggleButton.
    """

    def __init__(self, callback: Callable, **kwargs):
        super().__init__(**{**BUTTON_LAYOUT, **kwargs})
        self.callback = callback
        self.observe(self, names='value')

    def __call__(self, *ignored):
        self.callback()


class PlusMinusTool(ipw.HBox):
    def __init__(self, plus, minus):
        layout = {'width': '16px', 'padding': '0px'}
        self.callback = {'plus': plus.pop('callback'), 'minus': minus.pop('callback')}
        self._plus = ipw.Button(icon='plus', **{**{'layout': layout}, **plus})
        self._minus = ipw.Button(icon='minus', **{**{'layout': layout}, **minus})
        self._plus.on_click(self.plus)
        self._minus.on_click(self.minus)
        super().__init__([self._minus, self._plus])

    def plus(self, *ignored):
        self.callback['plus']()

    def minus(self, *ignored):
        self.callback['minus']()

    @property
    def disabled(self):
        return self._plus.disabled

    @disabled.setter
    def disabled(self, value):
        self._plus.disabled = value
        self._minus.disabled = value


class MultiToggleTool(ipw.VBox):
    """
    Create toggle buttons with a callback that is called when one of the buttons is
    toggled. In addition to ipywidgets ToggleButtons, when you click the button
    which is already selected, it resets the value to `None` (no button selected).

    Parameters
    ----------
    callback:
        The function that will be called when the one of the buttons is toggled.
    options:
        A list of values for the different buttons (one button per value will be
        created).
    icons:
        A list of icons (one icon per button).
    tooltips:
        A list of tooltips (one tooltip per button).
    descriptions:
        A list of descriptions (one description per button).
    value:
        If given, the value of the button with the corresponding option will be set to
        True.
    **kwargs:
        All other kwargs are forwarded to ipywidgets.ToggleButton.
    """

    def __init__(
        self,
        callback: Callable,
        options: list[str],
        icons: list[str] | None = None,
        tooltips: list[str] | None = None,
        descriptions: list[str] | None = None,
        value: str | None = None,
        **kwargs,
    ):
        self.callback = callback
        self._options = options
        self._buttons = {}
        for i, key in enumerate(self._options):
            tb = ipw.ToggleButton(
                icon=icons[i] if icons is not None else None,
                tooltip=tooltips[i] if tooltips is not None else None,
                description=descriptions[i] if descriptions is not None else '',
                value=key == value,
                **{**BUTTON_LAYOUT, **kwargs},
            )
            tb._option = key
            tb.observe(self, names='value')
            self._buttons[key] = tb
        self._lock = False
        super().__init__(list(self._buttons.values()))

    def __call__(self, change: dict):
        if self._lock:
            return
        key = change['owner']._option
        self._lock = True
        for name in set(self._buttons.keys()) - {key}:
            if self._buttons[name].value:
                self._buttons[name].value = False
        self._lock = False
        self.callback()

    @property
    def value(self) -> str | None:
        for b in self._buttons.values():
            if b.value:
                return b._option


HomeTool = partial(ButtonTool, icon='home', tooltip='Autoscale view')
"""Return home tool."""

LogxTool = partial(ToggleTool, description='logx', tooltip='Toggle X axis scale')
"""Toggle horizontal axis scale tool."""

LogyTool = partial(ToggleTool, description='logy', tooltip='Toggle Y axis scale')
"""Toggle vertical axis scale tool."""

LogNormTool = partial(
    ToggleTool, description='log', tooltip='Toggle colorscale normalization'
)
"""Toggle normalization scale tool."""

SaveTool = partial(ButtonTool, icon='save', tooltip='Save figure')
"""Save figure to png tool."""

CameraTool = partial(ButtonTool, icon='camera', tooltip='Autoscale view')
"""Tool for changing the position of the camera in a 3d scene."""

OutlineTool = partial(
    ToggleTool, value=True, icon='cube', tooltip='Toggle outline visibility'
)
"""Toggle outline visibility tool"""

AxesTool = partial(
    ToggleTool,
    value=True,
    icon='ruler-combined',
    style={'font_weight': 'bold'},
    tooltip='Toggle visibility of XYZ axes',
)
"""Toggle RGB axes helper visibility tool"""


class PanZoomTool(MultiToggleTool):
    """
    Tool to control the panning and zooming actions on a Matplotlib figure.

    Parameters
    ----------
    canvas:
        The canvas to act upon.
    value:
        Set the initially selected button. No button selected if ``None``.
    """

    def __init__(self, callback: Callable, value: bool | None = None, **kwargs):
        self._callback = callback
        super().__init__(
            callback=self._panzoom,
            options=['pan', 'zoom'],
            icons=['arrows', 'search-plus'],
            tooltips=['Pan', 'Zoom'],
            value=value,
            **kwargs,
        )

    def _panzoom(self):
        self._callback(self.value)


class ColorTool(ipw.HBox):
    """
    Tool for saving lines and controlling their color.
    It also comes with a close button to remove a line.

    Parameters
    ----------
    text:
        Text label for the tool.
    color:
        Initial color.
    """

    def __init__(self, text: str, color: str):
        layout = ipw.Layout(display="flex", justify_content="flex-end", width='150px')
        self.text = ipw.Label(value=text, layout=layout)
        self.color = ipw.ColorPicker(
            concise=True, value=color, description='', layout={'width': "30px"}
        )
        self.button = ipw.Button(icon='times', **BUTTON_LAYOUT)
        super().__init__([self.text, self.color, self.button])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from typing import Literal

import numpy as np
import scipp as sc

from ...core.utils import maybe_variable_to_number
from ..common import BoundingBox, axis_bounds


class Canvas:
    """
    Plotly-based canvas used to render 2D graphics.
    It provides a figure and some axes, as well as functions for controlling the zoom,
    panning, and the scale of the axes.

    Parameters
    ----------
    figsize:
        The width and height of the figure, in inches.
    title:
        The title to be placed above the figure.
    vmin:
        The minimum value for the vertical axis. If a number (without a unit) is
        supplied, it is assumed that the unit is the same as the current vertical axis
        unit.
    vmax:
        The maximum value for the vertical axis. If a number (without a unit) is
        supplied, it is assumed that the unit is the same as the current vertical axis
        unit.
    autoscale:
        The behavior of the axis limits. If ``auto``, the limits automatically
        adjusts every time the data changes. If ``grow``, the limits are allowed to
        grow with time but they do not shrink.
    """

    def __init__(
        self,
        figsize: tuple[float, float] | None = None,
        title: str | None = None,
        vmin: sc.Variable | float = None,
        vmax: sc.Variable | float = None,
        autoscale: Literal['auto', 'grow'] = 'auto',
        **ignored,
    ):
        # Note on the `**ignored`` keyword arguments: the figure which owns the canvas
        # creates both the canvas and an artist object (Line or Image). The figure
        # accepts keyword arguments, and has to somehow forward them to the canvas and
        # the artist. Since the figure has no detailed knowledge of the underlying
        # backend that implements the canvas, it cannot have specific arguments (such
        # as `layout` for specifying a Plotly layout).
        # Instead, we forward all the kwargs from the figure to both the canvas and the
        # artist, and filter out the artist kwargs with `**ignored`.

        import plotly.graph_objects as go

        self.fig = go.FigureWidget(
            layout={
                'modebar_remove': [
                    'zoom',
                    'pan',
                    'select',
                    'toImage',
                    'zoomIn',
                    'zoomOut',
                    'autoScale',
                    'resetScale',
                    'lasso2d',
                ],
                'margin': {'l': 0, 'r': 0, 't': 0 if title is None else 40, 'b': 0},
                'dragmode': False,
                'width': 600 if figsize is None else figsize[0],
                'height': 400 if figsize is None else figsize[1],
            }
        )
        self.figsize = figsize
        self._user_vmin = vmin
        self._user_vmax = vmax
        self.units = {}
        self.dims = {}
        self._own_axes = False
        self._autoscale = autoscale
        if title:
            self.title = title
        self._bbox = BoundingBox()

    def to_widget(self):
        return self.fig

    def autoscale(self):
        """
        Auto-scale the axes ranges to show all data in the canvas.
        """
        bbox = BoundingBox()
        lines = [trace for trace in self.fig.data if hasattr(trace, '_plopp_mask')]
        for line in lines:
            line_mask = sc.array(dims=['x'], values=line._plopp_mask)
            line_x = sc.DataArray(data=sc.array(dims=['x'], values=line.x))
            line_y = sc.array(dims=['x'], values=line.y)
            if line.error_y.array is not None:
                line_y = sc.concat(
                    [
                        line_y,
                        sc.array(dims=['x'], values=line.y + line.error_y.array),
                        sc.array(dims=['x'], values=line.y - line.error_y.array),
                    ],
                    dim='y',
                )
            line_y = sc.DataArray(data=line_y, masks={'mask': line_mask})
            line_bbox = BoundingBox(
                **{**axis_bounds(('xmin', 'xmax'), line_x, self.xscale, pad=True)},
                **{**axis_bounds(('ymin', 'ymax'), line_y, self.yscale, pad=True)},
            )
            bbox = bbox.union(line_bbox)

        self._bbox = self._bbox.union(bbox) if self._autoscale == 'grow' else bbox
        if self._user_vmin is not None:
            self._bbox.ymin = maybe_variable_to_number(
                self._user_vmin, unit=self.units.get('y')
            )
        if self._user_vmax is not None:
            self._bbox.ymax = maybe_variable_to_number(
                self._user_vmax, unit=self.units.get('y')
            )

        xrange = np.array([self._bbox.xmin, self._bbox.xmax])
        yrange = np.array([self._bbox.ymin, self._bbox.ymax])
        if self.xscale == 'log':
            xrange = np.log10(xrange)
        if self.yscale == 'log':
            yrange = np.log10(yrange)
        self.fig.update_layout(xaxis_range=xrange, yaxis_range=yrange)

    def save(self, filename: str):
        """
        Save the figure to file.
        The default directory for writing the file is the same as the
        directory where the script or notebook is running.

        Parameters
        ----------
        filename:
            Name of the output file. Possible file extensions are ``.jpg``, ``.png``,
            ``.svg``, ``.pdf``, and ``.html`.
        """
        ext = filename.split('.')[-1]
        if ext == 'html':
            self.fig.write_html(filename)
        else:
            self.fig.write_image(filename)

    def set_axes(self, dims, units, dtypes):
        """
        Set the axes dimensions and units.

        Parameters
        ----------
        dims:
            The dimensions of the data.
        units:
            The units of the data.
        dtypes:
            The data types of the data.
        """
        self.dims = dims
        self.units = units
        self.dtypes = dtypes

    @property
    def empty(self) -> bool:
        """
        Check if the canvas is empty.
        """
        return not self.dims

    @property
    def title(self) -> str:
        """
        Get or set the title of the plot.
        """
        return self.fig.layout.title.text

    @title.setter
    def title(self, text: str):
        layout = self.fig.layout
        if not text:
            layout.margin.t = 0
        elif layout.margin.t == 0:
            layout.margin.t = 40
        layout.title = text

    @property
    def xlabel(self) -> str:
        """
        Get or set the label of the x-axis.
        """
        return self.fig.layout.xaxis.title.text

    @xlabel.setter
    def xlabel(self, lab: str):
        self.fig.layout.xaxis.title = lab

    @property
    def ylabel(self) -> str:
        """
        Get or set the label of the y-axis.
        """
        return self.fig.layout.yaxis.title.text

    @ylabel.setter
    def ylabel(self, lab: str):
        self.fig.layout.yaxis.title = lab

    @property
    def xscale(self) -> Literal['linear', 'log']:
        """
        Get or set the scale of the x-axis ('linear' or 'log').
        """
        return self.fig.layout.xaxis.type or 'linear'

    @xscale.setter
    def xscale(self, scale: Literal['linear', 'log']):
        self.fig.update_xaxes(type=scale)

    @property
    def yscale(self) -> Literal['linear', 'log']:
        """
        Get or set the scale of the y-axis ('linear' or 'log').
        """
        return self.fig.layout.yaxis.type or 'linear'

    @yscale.setter
    def yscale(self, scale: Literal['linear', 'log']):
        self.fig.update_yaxes(type=scale)

    @property
    def xmin(self) -> float:
        """
        Get or set the lower (left) bound of the x-axis.
        """
        return self.fig.layout.xaxis.range[0]

    @xmin.setter
    def xmin(self, value: float):
        self.fig.layout.xaxis.range = [value, self.xmax]

    @property
    def xmax(self) -> float:
        """
        Get or set the upper (right) bound of the x-axis.
        """
        return self.fig.layout.xaxis.range[1]

    @xmax.setter
    def xmax(self, value: float):
        self.fig.layout.xaxis.range = [self.xmin, value]

    @property
    def xrange(self) -> tuple[float, float]:
        """
        Get or set the range/limits of the x-axis.
        """
        return self.fig.layout.xaxis.range

    @xrange.setter
    def xrange(self, value: tuple[float, float]):
        self.fig.layout.xaxis.range = value

    @property
    def ymin(self) -> float:
        """
        Get or set the lower (bottom) bound of the y-axis.
        """
        return self.fig.layout.yaxis.range[0]

    @ymin.setter
    def ymin(self, value: float):
        self.fig.layout.yaxis.range = [value, self.ymax]

    @property
    def ymax(self) -> float:
        """
        Get or set the upper (top) bound of the y-axis.
        """
        return self.fig.layout.yaxis.range[1]

    @ymax.setter
    def ymax(self, value: float):
        self.fig.layout.yaxis.range = [self.ymin, value]

    @property
    def yrange(self) -> tuple[float, float]:
        """
        Get or set the range/limits of the y-axis.
        """
        return self.fig.layout.yaxis.range

    @yrange.setter
    def yrange(self, value: tuple[float, float]):
        self.fig.layout.yaxis.range = value

    def reset_mode(self):
        """
        Reset the modebar mode to nothing, to disable all zoom/pan tools.
        """
        self.fig.update_layout(dragmode=False)

    def zoom(self):
        """
        Activate the underlying Plotly zoom tool.
        """
        self.fig.update_layout(dragmode='zoom')

    def pan(self):
        """
        Activate the underlying Plotly pan tool.
        """
        self.fig.update_layout(dragmode='pan')

    def panzoom(self, value: Literal['pan', 'zoom', None]):
        """
        Activate or deactivate the pan or zoom tool, depending on the input value.
        """
        if value == 'zoom':
            self.zoom()
        elif value == 'pan':
            self.pan()
        elif value is None:
            self.reset_mode()

    def download_figure(self):
        """
        Save the figure to a PNG file via a pop-up dialog.
        """
        self.fig.write_image('figure.png')

    def logx(self):
        """
        Toggle the scale between ``linear`` and ``log`` along the horizontal axis.
        """
        self.xscale = 'log' if self.xscale in ('linear', None) else 'linear'

    def logy(self):
        """
        Toggle the scale between ``linear`` and ``log`` along the vertical axis.
        """
        self.yscale = 'log' if self.yscale in ('linear', None) else 'linear'

    def finalize(self):
        """
        Finalize is called at the end of figure creation. Add any polishing operations
        here.
        """
        return



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from __future__ import annotations

import uuid
from itertools import chain
from typing import Any

from .view import View


def _no_replace_append(container: list[Node], item: Node, kind: str) -> None:
    """
    Append ``item`` to ``container`` if it is not already in it.
    """
    if item in container:
        tpe = 'View' if kind == 'view' else 'Node'
        raise ValueError(f"{tpe} {item} is already a {kind} in {container}.")
    container.append(item)


class Node:
    """
    A node that can have parent and children nodes, to create a graph.
    A node can be constructed from a callable ``func``, or a raw object. In the case
    of a raw object, a node wrapping the object will be created.

    Parameters
    ----------
    func:
        The callable that is called when data is requested from the node. This can also
        be a raw object, in which case this becomes a callable that returns the object.
    *parents:
        Positional arguments that represent the positional arguments of the function
        ``func``.
    *kwparents:
        Keyword arguments that represent the keyword arguments of the function ``func``.
    """

    def __init__(self, func: Any, *parents: Any, **kwparents: Any) -> None:
        func_is_callable = callable(func)
        self._input_value = None
        if func_is_callable:
            self.func = func
        else:
            self._input_value = func
            self.func = lambda: func
        self._id = uuid.uuid4().hex
        self.children = []
        self.views = []
        self.parents = [p if isinstance(p, Node) else Node(p) for p in parents]
        self.kwparents = {
            key: p if isinstance(p, Node) else Node(p) for key, p in kwparents.items()
        }
        self._data = None

        if func_is_callable:
            # Set automatic name from function name and arguments
            args_string = ', '.join(
                chain(
                    (f'arg_{i}' for i in range(len(self.parents))),
                    self.kwparents.keys(),
                )
            )
            fname = getattr(self.func, "__name__", str(self.func))
            self.name = f'{fname}({args_string})'
        else:
            val_str = f'={func!r}' if isinstance(func, int | float | str) else ""
            self.name = f'Input <{type(func).__name__}{val_str}>'

        # Attempt to set children after setting name in case error message is needed
        for parent in chain(self.parents, self.kwparents.values()):
            _no_replace_append(parent.children, self, 'child')

    def __call__(self) -> Any:
        return self.request_data()

    @property
    def id(self) -> str:
        """
        The unique uuid of the node. This differs from the ``name`` which can be any
        string.
        """
        return self._id

    @property
    def input_value(self) -> Any:
        """
        The input value of the node, if it is an input node.
        """
        return self._input_value

    @property
    def is_input_node(self) -> bool:
        """
        Whether the node is an input node.
        """
        return self._input_value is not None

    def remove(self) -> None:
        """
        Remove the node from the graph.
        This attempts to remove clear the list of parents of the node.
        The operation fails is the node has children, as removing it would leave the
        graph in an ill-defined state.
        """
        for child in self.children:
            if self in child.parents:
                child.parents.remove(self)
            child.kwparents = {
                key: parent for key, parent in child.kwparents.items() if parent != self
            }
            child._data = None
        self.children.clear()
        for view in self.views:
            del view.graph_nodes[self.id]
        for parent in chain(self.parents, self.kwparents.values()):
            parent.children.remove(self)
        self.views.clear()
        self.parents.clear()
        self.kwparents.clear()
        self._data = None

    def request_data(self) -> Any:
        """
        Request data from the node. This in turn requests data from all of the node's
        parents, and passes those results as arguments to the node's own ``func``.
        The result from calling the function is cached, to limit the number of times
        the graph is traversed.
        """
        if self._data is None:
            args = (parent.request_data() for parent in self.parents)
            kwargs = {
                key: parent.request_data() for key, parent in self.kwparents.items()
            }
            self._data = self.func(*args, **kwargs)
        return self._data

    def add_parents(self, *parents: Node) -> None:
        """
        Add one or more parents to the node.
        """
        for parent in parents:
            _no_replace_append(self.parents, parent, 'parent')
            _no_replace_append(parent.children, self, 'child')

    def add_kwparents(self, **parents: Node) -> None:
        """
        Add one or more keyword parents to the node.
        """
        for key, parent in parents.items():
            self.kwparents[key] = parent
            _no_replace_append(parent.children, self, 'child')

    def add_view(self, view: View) -> None:
        """
        Add a view to the node.
        """
        _no_replace_append(self.views, view, 'view')
        view.graph_nodes[self.id] = self

    def notify_children(self, message: Any) -> None:
        """
        Notify all of the node's children with ``message``.
        Receiving a notification also means that the local copy of the data is
        out-of-date, and it is thus reset.

        Parameters
        ----------
        message:
            The message to pass to the children.
        """
        self._data = None
        self.notify_views(message)
        for child in self.children:
            child.notify_children(message)

    def notify_views(self, message: Any) -> None:
        """
        Notify the node's views with ``message``.

        Parameters
        ----------
        message:
            The message to pass to the views.
        """
        for view in self.views:
            view.notify_view({"node_id": self.id, "message": message})

    def __eq__(self, other: Node | Any) -> bool:
        if not isinstance(other, Node):
            return False
        return self.id == other.id

    def __repr__(self) -> str:
        return f"Node(name={self.name})"

    def __add__(self, other: Node | Any) -> Node:
        return Node(lambda x, y: x + y, self, other)

    def __radd__(self, other: Node | Any) -> Node:
        return Node(lambda x, y: x + y, other, self)

    def __sub__(self, other: Node | Any) -> Node:
        return Node(lambda x, y: x - y, self, other)

    def __rsub__(self, other: Node | Any) -> Node:
        return Node(lambda x, y: x - y, other, self)

    def __mul__(self, other: Node | Any) -> Node:
        return Node(lambda x, y: x * y, self, other)

    def __rmul__(self, other: Node | Any) -> Node:
        return Node(lambda x, y: x * y, other, self)

    def __truediv__(self, other: Node | Any) -> Node:
        return Node(lambda x, y: x / y, self, other)

    def __rtruediv__(self, other: Node | Any) -> Node:
        return Node(lambda x, y: x / y, other, self)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from functools import partial
from typing import Literal

from scipp import Variable

from ..core.typing import FigureLike, PlottableMulti
from ..graphics import imagefigure, linefigure
from .common import input_to_nodes, preprocess, raise_multiple_inputs_for_2d_plot_error


def plot(
    obj: PlottableMulti,
    *,
    aspect: Literal['auto', 'equal'] = 'auto',
    cbar: bool = True,
    coords: list[str] | None = None,
    errorbars: bool = True,
    figsize: tuple[float, float] | None = None,
    grid: bool = False,
    ignore_size: bool = False,
    mask_color: str = 'black',
    norm: Literal['linear', 'log'] = 'linear',
    scale: dict[str, str] | None = None,
    title: str | None = None,
    vmin: Variable | float | None = None,
    vmax: Variable | float | None = None,
    autoscale: Literal['auto', 'grow'] = 'auto',
    legend: bool | tuple[float, float] = True,
    **kwargs,
) -> FigureLike:
    """Plot a Scipp object.

    Parameters
    ----------
    obj:
        The object to be plotted.
    aspect:
        Aspect ratio for the axes.
    cbar:
        Show colorbar in 2d plots if ``True``.
    coords:
        If supplied, use these coords instead of the input's dimension coordinates.
    errorbars:
        Show errorbars in 1d plots if ``True``.
    figsize:
        The width and height of the figure, in inches.
    grid:
        Show grid if ``True``.
    ignore_size:
        If ``True``, skip the check that prevents the rendering of very large data
        objects.
    mask_color:
        Color of masks in 1d plots.
    norm:
        Set to ``'log'`` for a logarithmic y-axis (1d plots) or logarithmic colorscale
        (2d plots).
    scale:
        Change axis scaling between ``log`` and ``linear``. For example, specify
        ``scale={'tof': 'log'}`` if you want log-scale for the ``tof`` dimension.
    title:
        The figure title.
    vmin:
        Lower bound for data to be displayed (y-axis for 1d plots, colorscale for
        2d plots).
    vmax:
        Upper bound for data to be displayed (y-axis for 1d plots, colorscale for
        2d plots).
    autoscale:
        The behavior of the axis (1d plots) or the color range limits (2d plots).
        If ``auto``, the limits automatically adjusts every time the data changes.
        If ``grow``, the limits are allowed to grow with time but they do not shrink.
    legend:
        Show legend if ``True``. If ``legend`` is a tuple, it should contain the
        ``(x, y)`` coordinates of the legend's anchor point in axes coordinates.
    **kwargs:
        All other kwargs are directly forwarded to Matplotlib, the underlying plotting
        library. The underlying functions called are the following:

        - 1d data with a non bin-edge coordinate: ``plot``
        - 1d data with a bin-edge coordinate: ``step``
        - 2d data: ``pcolormesh``

    Returns
    -------
    :
        A figure.
    """

    common_args = {
        'grid': grid,
        'norm': norm,
        'scale': scale,
        'title': title,
        'vmin': vmin,
        'vmax': vmax,
        'autoscale': autoscale,
        'figsize': figsize,
        **kwargs,
    }

    nodes = input_to_nodes(
        obj, processor=partial(preprocess, ignore_size=ignore_size, coords=coords)
    )

    ndims = set()
    for n in nodes:
        ndims.add(n().ndim)
    if len(ndims) > 1:
        raise ValueError(
            'All items given to the plot function must have the same '
            f'number of dimensions. Found dimensions {ndims}.'
        )
    ndim = ndims.pop()
    if ndim == 1:
        return linefigure(
            *nodes,
            errorbars=errorbars,
            mask_color=mask_color,
            legend=legend,
            **common_args,
        )
    elif ndim == 2:
        if len(nodes) > 1:
            raise_multiple_inputs_for_2d_plot_error(origin='plot')
        return imagefigure(
            *nodes,
            aspect=aspect,
            cbar=cbar,
            **common_args,
        )
    else:
        raise ValueError(
            'The plot function can only plot 1d and 2d data, got input '
            f'with {ndim} dimensions'
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from .inspector import inspector
from .plot import plot
from .scatter import scatter
from .scatter3d import scatter3d
from .slicer import slicer
from .superplot import superplot
from .xyplot import xyplot

__all__ = [
    'inspector',
    'plot',
    'scatter',
    'scatter3d',
    'slicer',
    'superplot',
    'xyplot',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import numpy as np
import pytest
import scipp as sc

from plopp.backends.matplotlib import MatplotlibBackend
from plopp.backends.matplotlib.interactive import InteractiveFig
from plopp.backends.matplotlib.static import StaticFig
from plopp.core import Node
from plopp.graphics.imageview import ImageView
from plopp.graphics.lineview import LineView


def test_create_static_fig1d():
    b = MatplotlibBackend()
    fig = b.figure1d(View=LineView)
    assert isinstance(fig, StaticFig)


@pytest.mark.usefixtures('_use_ipympl')
def test_create_interactive_fig1d():
    b = MatplotlibBackend()
    fig = b.figure1d(View=LineView)
    assert isinstance(fig, InteractiveFig)


def test_create_static_fig2d():
    b = MatplotlibBackend()
    fig = b.figure2d(View=ImageView)
    assert isinstance(fig, StaticFig)


@pytest.mark.usefixtures('_use_ipympl')
def test_create_interactive_fig2d():
    b = MatplotlibBackend()
    fig = b.figure2d(View=ImageView)
    assert isinstance(fig, InteractiveFig)


def test_datetime_compatibility_between_1d_and_2d_figures():
    b = MatplotlibBackend()
    # 2d data
    t = np.arange(
        np.datetime64('2017-03-16T20:58:17'), np.datetime64('2017-03-16T21:15:17'), 20
    )
    time = sc.array(dims=['time'], values=t)
    z = sc.arange('z', 50.0, unit='m')
    v = 10 * np.random.random(z.shape + time.shape)
    da2d = sc.DataArray(
        data=sc.array(dims=['z', 'time'], values=v), coords={'time': time, 'z': z}
    )
    fig = b.figure2d(ImageView, Node(da2d))
    assert len(fig.ax.lines) == 0
    assert len(fig.ax.collections) == 1

    # 1d data
    v = np.random.rand(time.sizes['time'])
    da1d = sc.DataArray(data=sc.array(dims=['time'], values=v), coords={'time': time})
    b.figure1d(LineView, Node(da1d), ax=fig.ax)
    assert len(fig.ax.lines) > 0
    assert len(fig.ax.collections) == 1



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import matplotlib
import matplotlib.pyplot as plt
import pytest


@pytest.fixture(autouse=True)
def _reset_mpl_defaults():
    matplotlib.rcdefaults()
    matplotlib.use('Agg')


@pytest.fixture(autouse=True)
def _close_figures():
    """
    Force closing all figures after each test case.
    Otherwise, the figures consume a lot of memory and matplotlib complains.
    """
    yield
    for fig in map(plt.figure, plt.get_fignums()):
        plt.close(fig)


@pytest.fixture()
def _use_ipympl():
    matplotlib.use('module://ipympl.backend_nbagg')


def pytest_sessionfinish(session, exitstatus):
    """
    When running no tests (e.g. in the noplotly tox env), pytest returns the exit code
    5, which causes the tox to fail. We use this to catch the exit status and convert
    it to 0.
    See https://github.com/pytest-dev/pytest/issues/2393#issuecomment-452634365
    """
    if exitstatus == 5:
        session.exitstatus = 0



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from __future__ import annotations

from typing import Protocol

import scipp as sc
from numpy import ndarray
from scipp.typing import VariableLike

from .node import Node


class VisibleDeprecationWarning(UserWarning):
    """Visible deprecation warning.

    By default, Python and in particular Jupyter will not show deprecation
    warnings, so this class can be used when a very visible warning is helpful.
    """


VisibleDeprecationWarning.__module__ = 'plopp'

Plottable = VariableLike | ndarray | Node

PlottableMulti = Plottable | dict[str, Plottable]


class CanvasLike(Protocol):
    def autoscale(self) -> None: ...

    def draw(self) -> None: ...

    def save(self) -> None: ...

    @property
    def empty(self) -> bool: ...

    @property
    def title(self) -> str: ...

    @title.setter
    def title(self, title: str) -> None: ...

    @property
    def xlabel(self) -> str: ...

    @xlabel.setter
    def xlabel(self, xlabel: str) -> None: ...

    @property
    def ylabel(self) -> str: ...

    @ylabel.setter
    def ylabel(self, ylabel: str) -> None: ...

    @property
    def xscale(self) -> str: ...

    @xscale.setter
    def xscale(self, xscale: str) -> None: ...

    @property
    def yscale(self) -> str: ...

    @yscale.setter
    def yscale(self, yscale: str) -> None: ...

    @property
    def xmin(self) -> float: ...

    @xmin.setter
    def xmin(self, xmin: float) -> None: ...

    @property
    def xmax(self) -> float: ...

    @xmax.setter
    def xmax(self, xmax: float) -> None: ...

    @property
    def ymin(self) -> float: ...

    @ymin.setter
    def ymin(self, ymin: float) -> None: ...

    @property
    def ymax(self) -> float: ...

    @ymax.setter
    def ymax(self, ymax: float) -> None: ...

    @property
    def xrange(self) -> tuple[float, float]: ...

    @xrange.setter
    def xrange(self, xrange: tuple[float, float]) -> None: ...

    @property
    def yrange(self) -> tuple[float, float]: ...

    @yrange.setter
    def yrange(self, yrange: tuple[float, float]) -> None: ...

    def logx(self) -> None: ...

    def logy(self) -> None: ...


class ArtistLike(Protocol):
    def update(self, new_values: sc.DataArray) -> None: ...


class FigureLike(Protocol):
    @property
    def canvas(self) -> CanvasLike: ...

    @property
    def artists(self) -> dict[str, ArtistLike]: ...

    @property
    def graph_nodes(self) -> dict[str, Node]: ...

    @property
    def id(self) -> str: ...

    def save(self, filename: str, **kwargs) -> None: ...

    def update(self, *args, **kwargs) -> None: ...

    def notify_view(self, *args, **kwargs) -> None: ...

    def copy(self, **kwargs) -> FigureLike: ...


__all__ = [
    'VisibleDeprecationWarning',
    'Plottable',
    'PlottableMulti',
    'CanvasLike',
    'ArtistLike',
    'FigureLike',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import pytest

import plopp as pp


@pytest.fixture(autouse=True, scope='module')
def _use_plotly():
    pp.backends['2d'] = 'plotly'
    yield
    pp.backends.reset()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from ipywidgets import HBox, VBox

from ...widgets import HBar, VBar, make_toolbar_canvas2d
from .figure import Figure


class InteractiveFig(Figure, VBox):
    """
    Create an interactive Matplotlib figure.
    """

    def __init__(self, View, *args, **kwargs):
        self.__init_figure__(View, *args, **kwargs)
        self.toolbar = make_toolbar_canvas2d(
            canvas=self._view.canvas,
            colormapper=getattr(self._view, 'colormapper', None),
        )
        self.left_bar = VBar([self.toolbar])
        self.right_bar = VBar()
        self.bottom_bar = HBar()
        self.top_bar = HBar()

        super().__init__(
            [
                self.top_bar,
                HBox([self.left_bar, self._view.canvas.to_widget(), self.right_bar]),
                self.bottom_bar,
            ]
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import uuid

import numpy as np
import scipp as sc

from ..common import make_line_data
from .canvas import Canvas


def _parse_dicts_in_kwargs(kwargs, name):
    out = {}
    for key, value in kwargs.items():
        if isinstance(value, dict):
            if name in value:
                out[key] = value[name]
        else:
            out[key] = value
    return out


class Line:
    """
    Artist to represent one-dimensional data.
    If the coordinate is bin centers, the line is (by default) a set of markers.
    If the coordinate is bin edges, the line is a step function.

    Parameters
    ----------
    canvas:
        The canvas that will display the line.
    data:
        The initial data to create the line from.
    number:
        The canvas keeps track of how many lines have been added to it. This number is
        used to set the color and marker parameters of the line.
    """

    def __init__(self, canvas: Canvas, data: sc.DataArray, number: int = 0, **kwargs):
        self._fig = canvas.fig
        self._data = data

        args = _parse_dicts_in_kwargs(kwargs, name=data.name)

        self._line = None
        self._mask = None
        self._error = None
        self._unit = None
        self.label = data.name
        self._dim = self._data.dim
        self._unit = self._data.unit
        self._coord = self._data.coords[self._dim]
        self._id = uuid.uuid4().hex

        self._make_line(
            data=make_line_data(data=self._data, dim=self._dim),
            number=number,
            **args,
        )

    def _make_line(
        self,
        data: dict,
        number: int,
        errorbars: bool = True,
        mask_color: str = 'black',
        mode='markers',
        marker=None,
        **kwargs,
    ):
        """
        Create either plot markers or a step function, depending on whether the data
        contains bin edges or not.

        Parameters
        ----------
        data:
            A dictionary containing data entries that have been pre-processed to be in
            a format that Matplotlib can directly use.
        number:
            The line number to set colors and marker style.
        errorbars:
            Show errorbars if ``True``.
        mask_color:
            The color to be used to represent the masks.
        **kwargs:
            The kwargs are forwarded to:

            - ``matplotlib.pyplot.plot`` for data with a non bin-edge coordinate
            - ``matplotlib.pyplot.step`` for data with a bin-edge coordinate
        """
        import plotly.graph_objects as go
        from plotly.colors import qualitative as plotly_colors

        default_colors = plotly_colors.Plotly
        default_line_style = {'color': default_colors[number % len(default_colors)]}
        default_marker_style = {'symbol': number % 53}  # Plotly has 52 marker styles

        line_shape = None

        if data["hist"]:
            line_shape = 'vh'
            mode = 'lines'

        marker_style = default_marker_style if marker is None else marker
        line_style = {**default_line_style, **kwargs}

        self._line = go.Scatter(
            x=np.asarray(data['values']['x']),
            y=np.asarray(data['values']['y']),
            name=self.label,
            mode=mode,
            marker=marker_style,
            line_shape=line_shape,
            line=line_style,
        )

        if errorbars and (data['stddevs'] is not None):
            self._error = go.Scatter(
                x=np.asarray(data['stddevs']['x']),
                y=np.asarray(data['stddevs']['y']),
                line=line_style,
                name=self.label,
                mode='markers',
                marker={'opacity': 0},
                error_y={'type': 'data', 'array': data['stddevs']['e']},
                showlegend=False,
            )

        marker_line_style = {'width': 3, 'color': mask_color}
        if 'line' in marker_style:
            marker_style['line'].update(marker_line_style)
        else:
            marker_style['line'] = marker_line_style
        if 'width' in line_style:
            line_style['width'] *= 5
        else:
            line_style['width'] = 5
        if data["hist"]:
            line_style['color'] = mask_color

        self._mask = go.Scatter(
            x=np.asarray(data['mask']['x']),
            y=np.asarray(data['mask']['y']),
            name=self.label,
            mode=mode,
            marker=marker_style,
            line_shape=line_shape,
            line=line_style,
            visible=data['mask']['visible'],
            showlegend=False,
        )

        # Below, we need to re-define the line because it seems that the Scatter trace
        # that ends up in the figure is a copy of the one above.
        # Plotly has no concept of zorder, so we need to add the traces in a specific
        # order
        if data["hist"]:
            self._fig.add_trace(self._mask)
            self._mask = self._fig.data[-1]
            self._fig.add_trace(self._line)
            self._line = self._fig.data[-1]
            if self._error is not None:
                self._fig.add_trace(self._error)
                self._error = self._fig.data[-1]
        else:
            self._fig.add_trace(self._line)
            self._line = self._fig.data[-1]
            if self._error is not None:
                self._fig.add_trace(self._error)
                self._error = self._fig.data[-1]
            self._fig.add_trace(self._mask)
            self._mask = self._fig.data[-1]
        self._line._plopp_id = self._id
        line_mask = ~np.isnan(data['mask']['y'])
        self._line._plopp_mask = line_mask
        self._mask._plopp_id = self._id
        if self._error is not None:
            self._error._plopp_id = self._id
            self._error._plopp_mask = line_mask[1:] if data["hist"] else line_mask

    def update(self, new_values: sc.DataArray):
        """
        Update the x and y positions of the data points from new data.

        Parameters
        ----------
        new_values:
            New data to update the line values, masks, errorbars from.
        """
        self._data = new_values
        new_values = make_line_data(data=self._data, dim=self._dim)
        line_mask = ~np.isnan(new_values['mask']['y'])

        with self._fig.batch_update():
            self._line.update(
                {'x': new_values['values']['x'], 'y': new_values['values']['y']}
            )
            self._line._plopp_mask = line_mask

            if (self._error is not None) and (new_values['stddevs'] is not None):
                self._error.update(
                    {
                        'x': new_values['stddevs']['x'],
                        'y': new_values['stddevs']['y'],
                        'error_y': {'array': new_values['stddevs']['e']},
                    }
                )
                self._error._plopp_mask = (
                    line_mask[1:] if new_values["hist"] else line_mask
                )

            if new_values['mask']['visible']:
                update = {'x': new_values['mask']['x'], 'y': new_values['mask']['y']}
                self._mask.update(update)
                self._mask.visible = True
            else:
                self._mask.visible = False

    def remove(self):
        """
        Remove the line, masks and errorbar artists from the canvas.
        """
        self._fig.data = [
            trace for trace in list(self._fig.data) if trace._plopp_id != self._id
        ]

    @property
    def color(self):
        """
        The line color.
        """
        return self._line.line.color

    @color.setter
    def color(self, val):
        self._line.line.color = val



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)


from functools import partial
from typing import Any, Literal

import ipywidgets as ipw
import numpy as np
import scipp as sc

from ..core import View, node
from ..utils import deprecated
from .debounce import debounce
from .style import BUTTON_LAYOUT
from .tools import PlusMinusTool


@deprecated("Use ``Clip3dTool`` instead.")
class Cut3dTool(ipw.HBox):
    """
    A tool that provides a slider to extract a plane of points in a three-dimensional
    scatter plot, and add it to the scene as an opaque cut. The slider controls the
    position of the slice. When the slider is dragged, the red outline of the cut is
    moved at the same time, while the actual point cloud gets updated less frequently
    using a debounce mechanism.

    The tool also has two buttons +/- to increase/decrease the thickness of the cut.

    .. deprecated:: v24.04.0

    Parameters
    ----------
    view:
        The 3d figure that contains the point clouds to be cut.
    limits:
        The spatial extent of the points in the 3d figure in the XYZ directions.
    direction:
        The direction normal to the slice.
    value:
        Set the cut to active upon creation if ``True``.
    color:
        Color of the cut's outline.
    linewidth:
        Width of the line delineating the outline.
    **kwargs:
        The kwargs are forwarded to the ToggleButton from ``ipywidgets``.
    """

    def __init__(
        self,
        view: View,
        limits: tuple[sc.Variable, sc.Variable, sc.Variable],
        direction: Literal['x', 'y', 'z'],
        value: bool = False,
        color: str = 'red',
        linewidth: float = 1.5,
        **kwargs,
    ):
        import pythreejs as p3

        self._limits = limits
        self._direction = direction
        axis = 'xyz'.index(self._direction)
        self._dim = self._limits[axis].dim
        self._unit = self._limits[axis].unit
        self._view = view

        w_axis = 2 if self._direction == 'x' else 0
        h_axis = 2 if self._direction == 'y' else 1
        width = (self._limits[w_axis][1] - self._limits[w_axis][0]).value
        height = (self._limits[h_axis][1] - self._limits[h_axis][0]).value

        self.outline = p3.LineSegments(
            geometry=p3.EdgesGeometry(
                p3.PlaneBufferGeometry(width=width, height=height)
            ),
            material=p3.LineBasicMaterial(color=color, linewidth=linewidth),
        )
        if self._direction == 'x':
            self.outline.rotateY(0.5 * np.pi)
        if self._direction == 'y':
            self.outline.rotateX(0.5 * np.pi)

        center = [var.mean().value for var in self._limits]
        self.outline.position = center
        self.outline.visible = value

        self.thickness = 0.01
        self._thickness_step = self.thickness * 0.25

        self.button = ipw.ToggleButton(value=value, **kwargs)
        self.slider = ipw.FloatSlider(
            min=limits[axis][0].value,
            max=limits[axis][1].value,
            value=center[axis],
            layout={'width': '150px', 'padding': '0px'},
            disabled=not value,
            readout=False,
        )
        self.slider.step = (self.slider.max - self.slider.min) * self.thickness * 0.5
        self.readout = ipw.FloatText(
            layout={'width': '70px'}, step=self.slider.step, disabled=not value
        )
        self.unit_label = ipw.Label(f'[{self._unit}]')
        ipw.jslink((self.slider, "value"), (self.readout, "value"))

        self.plus_minus = PlusMinusTool(
            plus={
                'callback': self.increase_thickness,
                'tooltip': 'Increase cut thickness',
                'disabled': not value,
            },
            minus={
                'callback': self.decrease_thickness,
                'tooltip': 'Decrease cut thickness',
                'disabled': not value,
            },
        )

        self.button.observe(self.toggle, names='value')
        self.slider.observe(self.move, names='value')
        self.slider.observe(self.update_cut, names='value')

        self._nodes = list(self._view.graph_nodes.values())
        self.select_nodes = {}

        super().__init__(
            [
                ipw.VBox([self.button, self.plus_minus]),
                ipw.VBox(
                    [self.slider, ipw.HBox([self.readout, self.unit_label])],
                    layout={'align_items': 'center'},
                ),
            ]
        )

    def toggle(self, change: dict[str, Any]):
        """
        Toggle the tool on and off.
        """
        self.outline.visible = change['new']
        for widget in (self.slider, self.plus_minus, self.readout):
            widget.disabled = not change['new']
        if change['new']:
            self._add_cut()
        else:
            self._remove_cut()

    def move(self, value: dict[str, Any]):
        """
        Move the outline of the cut according to new position given by the slider.
        """
        pos = list(self.outline.position)
        axis = 'xyz'.index(self._direction)
        pos[axis] = value['new']
        self.outline.position = pos
        self._remove_cut()

    def _add_cut(self):
        """
        Add a point cloud representing the points that intersect the cut plane.
        """

        def select(da, s):
            return da[s]

        for n in self._nodes:
            da = n.request_data()
            delta = sc.scalar(
                (self.slider.max - self.slider.min) * self.thickness, unit=self._unit
            )
            pos = sc.scalar(self.slider.value, unit=self._unit)
            selection = sc.abs(da.coords[self._dim] - pos) < delta
            if selection.sum().value > 0:
                self.select_nodes[n.id] = node(partial(select, s=selection))(da=n)
                self.select_nodes[n.id].add_view(self._view)
                self._view.update(
                    {self.select_nodes[n.id].id: self.select_nodes[n.id].request_data()}
                )

    def _remove_cut(self):
        """
        Remove a cut point the scene.
        """
        for n in self.select_nodes.values():
            self._view.remove(n.id)
            n.remove()
        self.select_nodes.clear()

    @debounce(0.3)
    def update_cut(self, _):
        """
        Update the position of the point cloud. This uses the debounce mechanism to
        avoid updating the cloud too often, and instead rely on simply moving the
        outline, which is cheap.
        """
        if self.value:
            self._add_cut()

    @property
    def value(self):
        """
        Return the state of the ToggleButton.
        """
        return self.button.value

    def increase_thickness(self):
        """
        Increase the thickness of the cut.
        """
        self.thickness = self.thickness + self._thickness_step
        self._remove_cut()
        self._add_cut()

    def decrease_thickness(self):
        """
        Decrease the thickness of the cut.
        """
        self.thickness = max(self.thickness - self._thickness_step, 0)
        self._remove_cut()
        self._add_cut()


@deprecated("Use ``ClippingPlanes`` instead.")
class TriCutTool(ipw.HBox):
    """
    A collection of :class:`Cut3dTool` to make spatial cuts in the X, Y, and Z
    directions on a three-dimensional scatter plot.

    .. deprecated:: v24.04.0

    Parameters
    ----------
    fig:
        The 3d figure that contains the point clouds to be cut.
    """

    def __init__(self, fig: View):
        self._fig = fig
        limits = self._fig.get_limits()
        self.cut_x = Cut3dTool(
            view=self._fig,
            direction='x',
            limits=limits,
            description='X',
            icon='cube',
            **BUTTON_LAYOUT,
        )
        self.cut_y = Cut3dTool(
            view=self._fig,
            direction='y',
            limits=limits,
            description='Y',
            icon='cube',
            **BUTTON_LAYOUT,
        )
        self.cut_z = Cut3dTool(
            view=self._fig,
            direction='z',
            limits=limits,
            description='Z',
            icon='cube',
            **BUTTON_LAYOUT,
        )

        self._fig.canvas.add(
            [self.cut_x.outline, self.cut_y.outline, self.cut_z.outline]
        )

        self.opacity = ipw.BoundedFloatText(
            min=0,
            max=0.5,
            step=0.01,
            disabled=True,
            value=0.03,
            style={'description_width': 'initial'},
            layout={'width': '50px', 'padding': '0px 0px 0px 0px'},
        )
        self.opacity.observe(self._set_opacity, names='value')

        self.cut_x.button.observe(self._toggle_opacity, names='value')
        self.cut_y.button.observe(self._toggle_opacity, names='value')
        self.cut_z.button.observe(self._toggle_opacity, names='value')

        space = ipw.HBox([], layout={'width': '10px'})
        super().__init__(
            [
                self.cut_x,
                space,
                self.cut_y,
                space,
                self.cut_z,
                space,
                ipw.VBox([ipw.Label('Opacity:'), self.opacity]),
            ]
        )
        self.layout.display = 'none'

    def _toggle_opacity(self, _):
        """
        If any cut is active, set the opacity of the original children (not the cuts) to
        a low value. If all cuts are inactive, set the opacity back to 1.
        """
        active_cut = any([self.cut_x.value, self.cut_y.value, self.cut_z.value])
        self.opacity.disabled = not active_cut
        opacity = self.opacity.value if active_cut else 1.0
        self._set_opacity({'new': opacity})

    def _set_opacity(self, change: dict[str, Any]):
        """
        Set the opacity of the original point clouds in the figure, not the cuts.
        """
        self._fig.set_opacity(change['new'])

    def toggle_visibility(self):
        """
        Toggle the visibility of the control buttons for making the cuts.
        """
        self.layout.display = None if self.layout.display == 'none' else 'none'



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import pytest
import scipp as sc

from plopp import Node
from plopp.data.testing import data_array
from plopp.graphics.imageview import ImageView


def test_empty():
    fig = ImageView()
    assert len(fig.artists) == 0


def test_update():
    fig = ImageView()
    assert len(fig.artists) == 0
    da = data_array(ndim=2)
    key = 'data2d'
    fig.update({key: da})
    assert sc.identical(fig.artists[key]._data, da)


def test_update_not_2d_raises():
    fig = ImageView()
    with pytest.raises(ValueError, match="Expected 2 dimension"):
        fig.update(data1d=data_array(ndim=1))
    with pytest.raises(ValueError, match="Expected 2 dimension"):
        fig.update(data3d=data_array(ndim=3))


def test_create_with_node():
    da = data_array(ndim=2)
    fig = ImageView(Node(da))
    assert len(fig.artists) == 1
    assert sc.identical(next(iter(fig.artists.values()))._data, da)


def test_create_with_bin_edges():
    da = data_array(ndim=2, binedges=True)
    fig = ImageView(Node(da))
    assert len(fig.artists) == 1
    assert sc.identical(next(iter(fig.artists.values()))._data, da)


def test_create_with_only_one_bin_edge_coord():
    da = data_array(ndim=2, binedges=True)
    da.coords['xx'] = sc.midpoints(da.coords['xx'])
    fig = ImageView(Node(da))
    assert len(fig.artists) == 1
    assert sc.identical(next(iter(fig.artists.values()))._data, da)


def test_log_norm():
    fig = ImageView(norm='log')
    assert fig.colormapper.norm == 'log'


def test_raises_for_new_data_with_incompatible_dimension():
    a = data_array(ndim=2)
    b = a.rename(xx='zz')
    with pytest.raises(KeyError):
        ImageView(Node(a), Node(b))


def test_raises_for_new_data_with_incompatible_unit():
    a = data_array(ndim=2)
    b = a * a
    with pytest.raises(sc.UnitError):
        ImageView(Node(a), Node(b))


def test_raises_for_new_data_with_incompatible_coord_unit():
    a = data_array(ndim=2)
    b = a.copy()
    b.coords['xx'] = a.coords['xx'] * a.coords['xx']
    with pytest.raises(sc.UnitError):
        ImageView(Node(a), Node(b))


def test_converts_new_data_units():
    a = data_array(ndim=2, unit='m')
    b = data_array(ndim=2, unit='cm')
    anode = Node(a)
    bnode = Node(b)
    fig = ImageView(anode, bnode)
    assert sc.identical(fig.artists[anode.id]._data, a)
    assert sc.identical(fig.artists[bnode.id]._data, b.to(unit='m'))


def test_converts_new_data_coordinate_units():
    a = data_array(ndim=2)
    b = data_array(ndim=2)
    b.coords['xx'].unit = 'cm'
    anode = Node(a)
    bnode = Node(b)
    fig = ImageView(anode, bnode)
    assert sc.identical(fig.artists[anode.id]._data, a)
    c = b.copy()
    c.coords['xx'] = c.coords['xx'].to(unit='m')
    assert sc.identical(fig.artists[bnode.id]._data, c)


def test_colorbar_label_has_correct_unit():
    da = data_array(ndim=2, unit='K')
    fig = ImageView(Node(da))
    assert fig.canvas.cblabel == '[K]'


def test_colorbar_label_has_correct_name():
    da = data_array(ndim=2, unit='K')
    name = 'My Experimental Data'
    da.name = name
    fig = ImageView(Node(da))
    assert fig.canvas.cblabel == name + ' [K]'


def test_colorbar_label_has_no_name_with_multiple_artists():
    a = data_array(ndim=2, unit='K')
    b = 3.3 * a
    a.name = 'A data'
    b.name = 'B data'
    fig = ImageView(Node(a), Node(b))
    assert fig.canvas.cblabel == '[K]'


def test_axis_label_with_transposed_2d_coord():
    a = sc.linspace('a', 0, 1, 10, unit='m')
    b = sc.linspace('b', 0, 2, 5, unit='s')
    da = sc.DataArray(a * b, coords={'a': a, 'b': b * a})
    fig = ImageView(Node(da))
    assert fig.canvas.xlabel == 'b [m*s]'
    da = sc.DataArray(a * b, coords={'a': a, 'b': a * b})
    fig2 = ImageView(Node(da))
    assert fig2.canvas.xlabel == fig.canvas.xlabel



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import numpy as np
import pytest
import scipp as sc

import plopp as pp


def test_xyplot_variable():
    x = sc.arange('time', 20.0, unit='s')
    y = sc.arange('time', 100.0, 120.0, unit='K')
    fig = pp.xyplot(x, y)
    assert fig.canvas.xlabel == 'time [s]'
    assert fig.canvas.ylabel == '[K]'


def test_xyplot_ndarray():
    N = 50
    x = np.arange(float(N))
    y = np.linspace(-44.0, 44.0, N)
    pp.xyplot(x, y)


def test_xyplot_list():
    x = [1, 2, 3, 4, 5]
    y = [6, 7, 2, 0, 1]
    pp.xyplot(x, y)


def test_xyplot_different_dims_raises():
    x = sc.arange('x', 20.0, unit='s')
    y = sc.arange('y', 100.0, 120.0, unit='K')
    with pytest.raises(sc.DimensionError, match='Dimensions of x and y must match'):
        pp.xyplot(x, y)


def test_xyplot_data_array_raises():
    x = sc.arange('x', 20.0, unit='s')
    y = pp.data.data1d()
    with pytest.raises(TypeError, match='Cannot convert input of type'):
        pp.xyplot(x, y)
    with pytest.raises(TypeError, match='Cannot convert input of type'):
        pp.xyplot(y, x)


def test_xyplot_2d_variable_raises():
    x = sc.arange('x', 50.0, unit='s')
    y = pp.data.data2d().data
    with pytest.raises(sc.DimensionError, match='Expected 1 dimensions, got 2'):
        pp.xyplot(x, y)
    with pytest.raises(sc.DimensionError, match='Expected 1 dimensions, got 2'):
        pp.xyplot(y, x)


def test_xyplot_variable_kwargs():
    x = sc.arange('time', 20.0, unit='s')
    y = sc.arange('time', 100.0, 120.0, unit='K')
    fig = pp.xyplot(x, y, color='red', vmin=102.0, vmax=115.0)
    assert np.allclose(fig.canvas.yrange, [102.0, 115.0])
    [line] = fig.artists.values()
    assert line.color == 'red'


def test_xyplot_bin_edges():
    x = sc.arange('time', 21.0, unit='s')
    y = sc.arange('time', 100.0, 120.0, unit='K')
    fig = pp.xyplot(x, y)
    [line] = fig.artists.values()
    assert len(line._line.get_xdata()) == 21


def test_xyplot_from_nodes():
    x = sc.arange('time', 20.0, unit='s')
    y = sc.arange('time', 100.0, 120.0, unit='K')
    pp.xyplot(pp.Node(x), y)
    pp.xyplot(x, pp.Node(y))
    pp.xyplot(pp.Node(x), pp.Node(y))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import uuid

import numpy as np
import scipp as sc

from ...core.utils import coord_as_bin_edges, merge_masks, repeat, scalar_to_string
from .canvas import Canvas


def _find_dim_of_2d_coord(coords):
    for xy, coord in coords.items():
        if coord['var'].ndim == 2:
            return (xy, coord['dim'])


def _get_dims_of_1d_and_2d_coords(coords):
    dim_2d = _find_dim_of_2d_coord(coords)
    if dim_2d is None:
        return None, None
    axis_1d = 'xy'.replace(dim_2d[0], '')
    dim_1d = (axis_1d, coords[axis_1d]['dim'])
    return dim_1d, dim_2d


def _maybe_repeat_values(data, dim_1d, dim_2d):
    if dim_2d is None:
        return data
    return repeat(data, dim=dim_1d[1], n=2)[dim_1d[1], :-1]


def _from_data_array_to_pcolormesh(data, coords, dim_1d, dim_2d):
    z = _maybe_repeat_values(data=data, dim_1d=dim_1d, dim_2d=dim_2d)
    if dim_2d is None:
        return coords['x'], coords['y'], z

    # Broadcast 1d coord to 2d and repeat along 1d dim
    # TODO: It may be more efficient here to first repeat and then broadcast, but
    # the current order is simpler in implementation.
    broadcasted_coord = repeat(
        sc.broadcast(
            coords[dim_1d[0]],
            sizes={**coords[dim_2d[0]].sizes, **coords[dim_1d[0]].sizes},
        ).transpose(data.dims),
        dim=dim_1d[1],
        n=2,
    )
    # Repeat 2d coord along 1d dim
    repeated_coord = repeat(coords[dim_2d[0]].transpose(data.dims), dim=dim_1d[1], n=2)
    out = {dim_1d[0]: broadcasted_coord[dim_1d[1], 1:-1], dim_2d[0]: repeated_coord}
    return out['x'], out['y'], z


class Image:
    """
    Artist to represent two-dimensional data.

    Parameters
    ----------
    canvas:
        The canvas that will display the line.
    data:
        The initial data to create the line from.
    shading:
        The shading to use for the ``pcolormesh``.
    rasterized:
        Rasterize the mesh/image if ``True``.
    **kwargs:
        Additional arguments are forwarded to Matplotlib's ``pcolormesh``.
    """

    def __init__(
        self,
        canvas: Canvas,
        data: sc.DataArray,
        shading: str = 'auto',
        rasterized: bool = True,
        **kwargs,
    ):
        self._canvas = canvas
        self._ax = self._canvas.ax
        self._data = data
        self._id = uuid.uuid4().hex
        # Because all keyword arguments from the figure are forwarded to both the canvas
        # and the line, we need to remove the arguments that belong to the canvas.
        kwargs.pop('ax', None)
        kwargs.pop('cax', None)
        # If the grid is visible on the axes, we need to set that on again after we
        # call pcolormesh, because that turns the grid off automatically.
        # See https://github.com/matplotlib/matplotlib/issues/15600.
        need_grid = self._ax.xaxis.get_gridlines()[0].get_visible()

        to_dim_search = {}
        string_labels = {}
        bin_edge_coords = {}
        self._data_with_bin_edges = sc.DataArray(data=self._data.data)
        for i, k in enumerate('yx'):
            to_dim_search[k] = {
                'dim': self._data.dims[i],
                'var': self._data.coords[self._data.dims[i]],
            }
            bin_edge_coords[k] = coord_as_bin_edges(self._data, self._data.dims[i])
            self._data_with_bin_edges.coords[self._data.dims[i]] = bin_edge_coords[k]
            if self._data.coords[self._data.dims[i]].dtype == str:
                string_labels[k] = self._data.coords[self._data.dims[i]]

        self._dim_1d, self._dim_2d = _get_dims_of_1d_and_2d_coords(to_dim_search)
        self._mesh = None

        x, y, z = _from_data_array_to_pcolormesh(
            data=self._data.data,
            coords=bin_edge_coords,
            dim_1d=self._dim_1d,
            dim_2d=self._dim_2d,
        )
        self._mesh = self._ax.pcolormesh(
            x.values,
            y.values,
            z.values,
            shading=shading,
            rasterized=rasterized,
            **kwargs,
        )
        self._mesh.set_array(None)

        for xy, var in string_labels.items():
            getattr(self._ax, f'set_{xy}ticks')(np.arange(float(var.shape[0])))
            getattr(self._ax, f'set_{xy}ticklabels')(var.values)

        if need_grid:
            self._ax.grid(True)

        self._canvas.register_format_coord(self.format_coord)

    @property
    def data(self):
        """
        Get the Mesh's data in a form that may have been tweaked, compared to the
        original data, in the case of a two-dimensional coordinate.
        """
        out = sc.DataArray(
            data=_maybe_repeat_values(
                data=self._data.data, dim_1d=self._dim_1d, dim_2d=self._dim_2d
            )
        )
        if self._data.masks:
            out.masks['one_mask'] = _maybe_repeat_values(
                data=sc.broadcast(
                    merge_masks(self._data.masks),
                    dims=self._data.dims,
                    shape=self._data.shape,
                ),
                dim_1d=self._dim_1d,
                dim_2d=self._dim_2d,
            )
        return out

    def set_colors(self, rgba: np.ndarray):
        """
        Set the mesh's rgba colors:

        Parameters
        ----------
        rgba:
            The array of rgba colors.
        """
        self._mesh.set_facecolors(rgba.reshape(np.prod(rgba.shape[:-1]), 4))
        self._canvas.draw()

    def update(self, new_values: sc.DataArray):
        """
        Update image array with new values.

        Parameters
        ----------
        new_values:
            New data to update the mesh values from.
        """
        self._data = new_values
        self._data_with_bin_edges.data = new_values.data

    def format_coord(
        self, xslice: tuple[str, sc.Variable], yslice: tuple[str, sc.Variable]
    ) -> str:
        """
        Format the coordinates of the mouse pointer to show the value of the
        data at that point.

        Parameters
        ----------
        xslice:
            Dimension and x coordinate of the mouse pointer, as slice parameters.
        yslice:
            Dimension and y coordinate of the mouse pointer, as slice parameters.
        """
        try:
            val = self._data_with_bin_edges[yslice][xslice]
            prefix = self._data.name
            if prefix:
                prefix += ': '
            return prefix + scalar_to_string(val)
        except IndexError:
            return None



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import scipp as sc

import plopp as pp
from plopp.data.testing import data_array
from plopp.data.testing import scatter as scatter_data


def test_scatter_simple():
    pp.scatter(scatter_data())


def test_scatter_styling():
    pp.scatter(scatter_data(), color='r', marker='P')


def test_scatter_other_coords():
    a = scatter_data()
    a.coords['x2'] = a.coords['x'] ** 2
    pp.scatter(a, x='x2', y='z')


def test_scatter_two_inputs():
    a = scatter_data()
    b = scatter_data(seed=2) * 10.0
    b.coords['x'] += sc.scalar(50.0, unit='m')
    pp.scatter({'a': a, 'b': b})


def test_scatter_two_inputs_color():
    a = scatter_data()
    b = scatter_data(seed=2) * 10.0
    b.coords['x'] += sc.scalar(50.0, unit='m')
    pp.scatter({'a': a, 'b': b}, color={'a': 'k', 'b': 'g'})


def test_scatter_with_colorbar():
    scat = pp.scatter(scatter_data(), cbar=True)
    assert scat._view.colormapper is not None


def test_scatter_with_cmap():
    name = 'magma'
    scat = pp.scatter(scatter_data(), cbar=True, cmap=name)
    assert scat._view.colormapper.cmap.name == name


def test_scatter_with_size():
    a = scatter_data()
    a.coords['s'] = sc.abs(a.coords['x']) * 5
    pp.scatter(a, size='s')


def test_scatter_with_size_and_cbar():
    a = scatter_data()
    a.coords['s'] = sc.abs(a.coords['x']) * 5
    pp.scatter(a, size='s', cbar=True)


def test_scatter_with_masks():
    a = scatter_data()
    a.masks['m'] = a.coords['x'] > sc.scalar(10, unit='m')
    pp.scatter(a)


def test_scatter_flattens_2d_data():
    pp.scatter(data_array(ndim=2), x='xx', y='yy', cbar=True)


def test_scatter_with_norm():
    a = scatter_data()
    scat = pp.scatter(a, cbar=True, norm='linear')
    assert scat._view.colormapper.norm == 'linear'
    scat = pp.scatter(a, cbar=True, norm='log')
    assert scat._view.colormapper.norm == 'log'


def test_scatter_log_axes():
    a = scatter_data()
    scat = pp.scatter(a)
    assert scat._view.canvas.xscale == 'linear'
    assert scat._view.canvas.yscale == 'linear'
    scat = pp.scatter(a, scale={'x': 'log', 'y': 'log'})
    assert scat._view.canvas.xscale == 'log'
    assert scat._view.canvas.yscale == 'log'



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

# from typing import Literal

from .. import backends


def tiled(nrows: int, ncols: int, **kwargs):
    """
    A tiled figure.

    .. versionadded:: 23.08.0

    Parameters
    ----------
    nrows:
        Number of rows.
    ncols:
        Number of columns.
    **kwargs:
        Additional arguments passed to :class:`Tiled`.

    Examples
    --------
    Create a tiled figure with two plots stacked vertically:

      >>> da1 = pp.data.data1d()
      >>> da2 = pp.data.data2d()
      >>> tiled = pp.tiled(2, 1)
      >>> tiled[0] = da1.plot()
      >>> tiled[1] = da2.plot()

    Create a tiled 2x2 figure:

      >>> da1 = pp.data.data1d()
      >>> da2 = pp.data.data2d()
      >>> da3 = pp.data.data2d()
      >>> da4 = pp.data.data1d()
      >>> tiled = pp.tiled(2, 2)
      >>> tiled[0, 0] = da1.plot()
      >>> tiled[0, 1] = da2.plot()
      >>> tiled[1, 0] = da3.plot()
      >>> tiled[1, 1] = da4.plot()

    Create a tiled figure with two figures side by side and the first is twice as wide:

      >>> da1 = pp.data.data1d()
      >>> da2 = pp.data.data2d()
      >>> tiled = pp.tiled(1, 3)
      >>> tiled[0, :2] = da1.plot()
      >>> tiled[0, 2] = da2.plot()

    """
    return backends.tiled(nrows=nrows, ncols=ncols, **kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import pytest

import plopp as pp
from plopp.backends.plotly.figure import Figure
from plopp.data.testing import data_array
from plopp.graphics.lineview import LineView

pytest.importorskip("plotly")


def test_creation():
    da = data_array(ndim=1)
    fig = Figure(LineView, pp.Node(da))
    assert fig.canvas.xlabel == f'xx [{da.coords["xx"].unit}]'
    assert fig.canvas.ylabel == f'[{da.unit}]'


def test_logx_1d_toolbar_button():
    da = data_array(ndim=1)
    fig = Figure(LineView, pp.Node(da), scale={'xx': 'log'})
    assert fig.toolbar['logx'].value


def test_logy_1d_toolbar_button():
    da = data_array(ndim=1)
    fig = Figure(LineView, pp.Node(da), norm='log')
    assert fig.toolbar['logy'].value



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

from typing import Literal

import scipp as sc

from .. import backends
from ..core.typing import FigureLike
from .colormapper import ColorMapper
from .graphicalview import GraphicalView


class ScatterView(GraphicalView):
    """ """

    def __init__(
        self,
        *nodes,
        x: str = 'x',
        y: str = 'y',
        size: str | None = None,
        norm: Literal['linear', 'log'] = 'linear',
        vmin: sc.Variable | float | None = None,
        vmax: sc.Variable | float | None = None,
        autoscale: Literal['auto', 'grow'] = 'auto',
        scale: dict[str, str] | None = None,
        mask_color: str = 'black',
        aspect: Literal['auto', 'equal'] = 'auto',
        grid: bool = False,
        title: str | None = None,
        figsize: tuple[float, float] | None = None,
        format: Literal['svg', 'png'] | None = None,
        legend: bool | tuple[float, float] = True,
        cmap: str = 'viridis',
        mask_cmap: str = 'gray',
        cbar: bool = False,
        **kwargs,
    ):
        super().__init__(*nodes)

        self._dims = {'x': x, 'y': y}
        self._ndim = 1
        self._size = size
        self._cbar = cbar
        self._scale = {} if scale is None else scale
        self._mask_color = mask_color
        self._kwargs = kwargs
        self._repr_format = format
        self.canvas = backends.canvas2d(
            cbar=self._cbar,
            aspect=aspect,
            grid=grid,
            figsize=figsize,
            title=title,
            vmin=vmin,
            vmax=vmax,
            autoscale=autoscale,
            legend=legend,
            **kwargs,
        )
        if self._cbar:
            self.colormapper = ColorMapper(
                cmap=cmap,
                cbar=True,
                mask_cmap=mask_cmap,
                norm=norm,
                vmin=vmin,
                vmax=vmax,
                autoscale=autoscale,
                canvas=self.canvas,
            )
        else:
            self.colormapper = None

        self.render()
        self.canvas.finalize()

    def make_artist(self, new_values):
        return backends.scatter(
            canvas=self.canvas,
            data=new_values,
            x=self._dims['x'],
            y=self._dims['y'],
            size=self._size,
            number=len(self.artists),
            mask_color=self._mask_color,
            cbar=self._cbar,
            **self._kwargs,
        )


def scatterfigure(*args, **kwargs) -> FigureLike:
    """
    Create a figure to represent scatter data from one or more graph node(s).

    .. versionadded:: 24.04.0

    Examples
    --------
    Create an input node and attach a ``scatterfigure`` as a view:

      >>> da = pp.data.scatter()
      >>> in_node = pp.Node(da)
      >>> fig = pp.scatterfigure(in_node)

    A scatter figure with a color bar (using the data values for the color scale):

      >>> da = pp.data.scatter()
      >>> in_node = pp.Node(da)
      >>> fig = pp.scatterfigure(in_node, cbar=True)
    """

    return backends.figure2d(ScatterView, *args, **kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from io import BytesIO
from typing import Literal

import matplotlib as mpl
from matplotlib.pyplot import Figure, _get_backend_mod

from ...core.typing import FigureLike


def fig_to_bytes(fig: Figure, form: Literal['png', 'svg'] = 'png') -> bytes:
    """
    Convert a Matplotlib figure to png (default) or svg bytes.

    Parameters
    ----------
    fig:
        The figure to be converted.
    form:
        The format to use.
    """
    buf = BytesIO()
    fig.savefig(buf, format=form, bbox_inches='tight')
    buf.seek(0)
    return buf.getvalue()


def is_interactive_backend() -> bool:
    """
    Return ``True`` if the current backend used by Matplotlib is the widget/ipympl
    backend.
    """
    backend = mpl.get_backend()
    return any(x in backend for x in ("ipympl", "widget"))


def make_figure(*args, **kwargs) -> Figure:
    """
    Create a new figure.

    If we use ``plt.figure()`` directly, the figures auto-show in the notebooks.
    We want to display the figures when the figure repr is requested.

    When using the static backend, we can return the ``plt.Figure`` (note the uppercase
    F) directly.
    When using the interactive backend, we need to do more work. The ``plt.Figure``
    will not have a toolbar nor will it be interactive, as opposed to what
    ``plt.figure`` returns. We therefore copy the minimal required code inside the
    ``plt.figure`` function which creates a figure manager (which is apparently what
    creates the toolbar and makes the figure interactive).
    """
    if not is_interactive_backend():
        return Figure(*args, **kwargs)
    backend = _get_backend_mod()
    manager = backend.new_figure_manager(1, *args, FigureClass=Figure, **kwargs)
    return manager.canvas.figure


def make_legend(leg: bool | tuple[float, float]):
    """
    Create a dict of arguments to be used in the legend creation.
    """
    leg_args = {}
    if isinstance(leg, list | tuple):
        leg_args = {'loc': leg}
    elif not isinstance(leg, bool):
        raise TypeError(f"Legend must be a bool, tuple, or a list, not {type(leg)}")
    return leg_args


def require_interactive_backend(func: str):
    """
    Raise an error if the current backend in use is non-interactive.
    """
    if not is_interactive_backend():
        raise RuntimeError(
            f"The {func} can only be used with the interactive widget "
            "backend. Use `%matplotlib widget` at the start of your "
            "notebook."
        )


def _running_in_jupyter() -> bool:
    """
    Detect whether Python is running in Jupyter.

    Note that this includes not only Jupyter notebooks
    but also Jupyter console and qtconsole.
    """
    try:
        import ipykernel.zmqshell
        from IPython import get_ipython
    except ImportError:
        # Cannot be Jupyter if IPython is not installed.
        return False

    return isinstance(get_ipython(), ipykernel.zmqshell.ZMQInteractiveShell)


def is_sphinx_build() -> bool:
    """
    Returns ``True`` if we are running inside a sphinx documentation build.
    """
    if not _running_in_jupyter():
        return False
    from IPython import get_ipython

    ipy = get_ipython()
    cfg = ipy.config
    meta = cfg["Session"]["metadata"]
    if hasattr(meta, "to_dict"):
        meta = meta.to_dict()
    return meta.get("scipp_sphinx_build", False)


def copy_figure(fig: FigureLike, **kwargs) -> FigureLike:
    out = fig.__class__(
        fig._view.__class__,
        *fig._args,
        **{**fig._kwargs, **kwargs},
    )
    for prop in ('xrange', 'yrange', 'xscale', 'yscale', 'title', 'grid'):
        setattr(out.canvas, prop, getattr(fig.canvas, prop))
    return out


def parse_dicts_in_kwargs(kwargs, name):
    out = {}
    for key, value in kwargs.items():
        if isinstance(value, dict):
            if name in value:
                out[key] = value[name]
        else:
            out[key] = value
    return out



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import numpy as np
import pytest
import scipp as sc

from plopp.core.limits import find_limits, fix_empty_range


def test_find_limits():
    da = sc.DataArray(data=sc.arange('x', 11.0, unit='m'))
    lims = find_limits(da)
    assert sc.identical(lims[0], sc.scalar(0.0, unit='m'))
    assert sc.identical(lims[1], sc.scalar(10.0, unit='m'))


def test_find_limits_log():
    da = sc.DataArray(data=sc.arange('x', 11.0, unit='m'))
    lims = find_limits(da, scale='log')
    assert sc.identical(lims[0], sc.scalar(1.0, unit='m'))
    assert sc.identical(lims[1], sc.scalar(10.0, unit='m'))


def test_find_limits_log_int():
    da = sc.DataArray(data=sc.arange('x', 11, unit='m'))
    lims = find_limits(da, scale='log')
    assert sc.identical(lims[0], sc.scalar(1, unit='m'))
    assert sc.identical(lims[1], sc.scalar(10, unit='m'))


def test_find_limits_with_nan():
    da = sc.DataArray(data=sc.arange('x', 11.0, unit='m'))
    da.values[5] = np.nan
    lims = find_limits(da)
    assert sc.identical(lims[0], sc.scalar(0.0, unit='m'))
    assert sc.identical(lims[1], sc.scalar(10.0, unit='m'))


def test_find_limits_with_inf():
    da = sc.DataArray(data=sc.arange('x', 11.0, unit='m'))
    da.values[5] = np.inf
    lims = find_limits(da)
    assert sc.identical(lims[0], sc.scalar(0.0, unit='m'))
    assert sc.identical(lims[1], sc.scalar(10.0, unit='m'))


def test_find_limits_with_ninf():
    da = sc.DataArray(data=sc.arange('x', 11.0, unit='m'))
    da.values[5] = -np.inf
    lims = find_limits(da)
    assert sc.identical(lims[0], sc.scalar(0.0, unit='m'))
    assert sc.identical(lims[1], sc.scalar(10.0, unit='m'))


def test_find_limits_no_finite_values_raises():
    da = sc.DataArray(
        data=sc.array(dims=['x'], values=[np.nan, np.inf, -np.inf, np.nan], unit='m')
    )
    with pytest.raises(ValueError, match="No finite values were found in array"):
        _ = find_limits(da)


def test_find_limits_all_zeros():
    da = sc.DataArray(data=sc.zeros(sizes={'x': 5}, unit='s'))
    lims = find_limits(da)
    assert sc.identical(lims[0], sc.scalar(0.0, unit='s'))
    assert sc.identical(lims[1], sc.scalar(0.0, unit='s'))


def test_find_limits_all_zeros_log_uses_default_positive_values():
    da = sc.DataArray(data=sc.zeros(sizes={'x': 5}, unit='s'))
    lims = find_limits(da, scale='log')
    zero = sc.scalar(0.0, unit='s')
    assert lims[0] > zero
    assert lims[1] > zero
    assert lims[0] < lims[1]


def test_find_limits_ignores_masks():
    x = sc.arange('x', 11.0, unit='m')
    da = sc.DataArray(data=x, masks={'mask': x > sc.scalar(5.0, unit='m')})
    lims = find_limits(da)
    assert sc.identical(lims[0], sc.scalar(0.0, unit='m'))
    assert sc.identical(lims[1], sc.scalar(5.0, unit='m'))


def test_find_limits_ignores_masks_log():
    x = sc.arange('x', 11.0, unit='m')
    da = sc.DataArray(data=x, masks={'mask': x < sc.scalar(3.0, unit='m')})
    lims = find_limits(da)
    assert sc.identical(lims[0], sc.scalar(3.0, unit='m'))
    assert sc.identical(lims[1], sc.scalar(10.0, unit='m'))


def test_find_limits_with_padding():
    da = sc.DataArray(data=sc.arange('x', 11.0, unit='m'))
    lims = find_limits(da, pad=True)
    assert lims[0] < sc.scalar(0.0, unit='m')
    assert lims[1] > sc.scalar(10.0, unit='m')


def test_find_limits_with_padding_log():
    da = sc.DataArray(data=sc.arange('x', 11.0, unit='m'))
    lims = find_limits(da, scale='log', pad=True)
    assert lims[0] < sc.scalar(1.0, unit='m')
    assert lims[0] > sc.scalar(0.0, unit='m')
    assert lims[1] > sc.scalar(10.0, unit='m')


def test_find_limits_with_strings():
    da = sc.DataArray(
        data=sc.array(dims=['x'], values=['a', 'b', 'c', 'd', 'e'], unit='K')
    )
    lims = find_limits(da)
    assert sc.identical(lims[0], sc.scalar(0.0, unit='K'))
    assert sc.identical(lims[1], sc.scalar(4.0, unit='K'))


def test_find_limits_datetime():
    time = np.arange(
        np.datetime64('2017-03-16T20:58:17'), np.datetime64('2017-03-16T21:15:17'), 20
    )
    da = sc.DataArray(data=sc.array(dims=['time'], values=time))
    lims = find_limits(da)
    assert sc.identical(lims[0], sc.scalar(time[0], unit='s'))
    assert sc.identical(lims[1], sc.scalar(time[-1], unit='s'))


def test_fix_empty_range():
    a = sc.scalar(5.0, unit='m')
    b = sc.scalar(10.0, unit='m')
    lims = fix_empty_range((a, b))
    assert sc.identical(lims[0], a)
    assert sc.identical(lims[1], b)


def test_fix_empty_range_same_values():
    a = sc.scalar(5.0, unit='m')
    lims = fix_empty_range((a, a))
    assert sc.identical(lims[0], sc.scalar(2.5, unit='m'))
    assert sc.identical(lims[1], sc.scalar(7.5, unit='m'))
    lims = fix_empty_range((-a, -a))
    assert sc.identical(lims[0], sc.scalar(-7.5, unit='m'))
    assert sc.identical(lims[1], sc.scalar(-2.5, unit='m'))


def test_fix_empty_range_zeros_values():
    a = sc.scalar(0.0, unit='m')
    lims = fix_empty_range((a, a))
    assert sc.identical(lims[0], sc.scalar(-0.5, unit='m'))
    assert sc.identical(lims[1], sc.scalar(0.5, unit='m'))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from copy import copy

import pytest
import scipp as sc

from plopp import Camera
from plopp.backends.pythreejs.canvas import Canvas
from plopp.backends.pythreejs.outline import Outline


def _make_limits():
    return (
        sc.array(dims=['x'], values=[1.5, 3.5], unit='m'),
        sc.array(dims=['y'], values=[-5.0, 5.0], unit='m'),
        sc.array(dims=['z'], values=[-0.1, 17.0], unit='m'),
    )


def _assert_pos(expected, canvas, key):
    assert (
        expected == canvas.camera.position
        if key == 'position'
        else canvas.controls.target
    )


def test_creation():
    canvas = Canvas(figsize=(700, 450))
    assert canvas.renderer.width == 700
    assert canvas.renderer.height == 450
    assert all(
        [canvas.camera, canvas.scene, canvas.controls, canvas.renderer, canvas.axes_3d]
    )
    assert canvas.outline is None


def test_make_outline():
    canvas = Canvas()
    assert canvas.outline is None
    canvas.make_outline(limits=_make_limits())
    assert isinstance(canvas.outline, Outline)
    assert canvas.outline.visible


def test_camera_home():
    canvas = Canvas()
    canvas._update_camera(limits=_make_limits())
    original = copy(canvas.camera.position)
    canvas.camera.position = [10, 20, 30]
    canvas.home()
    assert canvas.camera.position == original


def test_camera_x():
    canvas = Canvas()
    canvas._update_camera(limits=_make_limits())
    center = copy(canvas.camera_backup["center"])
    canvas.camera.position = [10, 20, 30]
    canvas.camera_x_normal()
    assert canvas.camera.position[0] < center[0]
    assert canvas.camera.position[1] == center[1]
    assert canvas.camera.position[2] == center[2]
    canvas.camera_x_normal()
    assert canvas.camera.position[0] > center[0]
    assert canvas.camera.position[1] == center[1]
    assert canvas.camera.position[2] == center[2]


def test_camera_y():
    canvas = Canvas()
    canvas._update_camera(limits=_make_limits())
    center = copy(canvas.camera_backup["center"])
    canvas.camera.position = [10, 20, 30]
    canvas.camera_y_normal()
    assert canvas.camera.position[0] == center[0]
    assert canvas.camera.position[1] < center[1]
    assert canvas.camera.position[2] == center[2]
    canvas.camera_y_normal()
    assert canvas.camera.position[0] == center[0]
    assert canvas.camera.position[1] > center[1]
    assert canvas.camera.position[2] == center[2]


def test_camera_z():
    canvas = Canvas()
    canvas._update_camera(limits=_make_limits())
    center = copy(canvas.camera_backup["center"])
    canvas.camera.position = [10, 20, 30]
    canvas.camera_z_normal()
    assert canvas.camera.position[0] == center[0]
    assert canvas.camera.position[1] == center[1]
    assert canvas.camera.position[2] < center[2]
    canvas.camera_z_normal()
    assert canvas.camera.position[0] == center[0]
    assert canvas.camera.position[1] == center[1]
    assert canvas.camera.position[2] > center[2]


@pytest.mark.parametrize('key', ['position', 'look_at'])
def test_camera_user_tuple_of_floats(key):
    pos = (1.5, 22.0, -3.0)
    canvas = Canvas(camera=Camera(**{key: pos}))
    canvas._update_camera(limits=_make_limits())
    _assert_pos(pos, canvas, key)
    setattr(canvas.camera, key, (10, 20, 30))
    canvas.home()
    _assert_pos(pos, canvas, key)


@pytest.mark.parametrize('key', ['position', 'look_at'])
def test_camera_user_vector(key):
    pos = (0, 1, 2)
    vec = sc.vector(pos, unit='m')
    canvas = Canvas(camera=Camera(**{key: vec}))
    canvas.units.update(x='m', y='m', z='m')
    canvas._update_camera(limits=_make_limits())
    _assert_pos(pos, canvas, key)


@pytest.mark.parametrize('key', ['position', 'look_at'])
def test_camera_user_vector_unit_conversion(key):
    vec = sc.vector([0, 1, 2], unit='cm')
    canvas = Canvas(camera=Camera(**{key: vec}))
    canvas.units.update(x='m', y='m', z='m')
    canvas._update_camera(limits=_make_limits())
    _assert_pos((0, 0.01, 0.02), canvas, key)


@pytest.mark.parametrize('key', ['position', 'look_at'])
def test_camera_user_vector_bad_unit_raises(key):
    vec = sc.vector([0, 1, 2], unit='s')
    canvas = Canvas(camera=Camera(**{key: vec}))
    canvas.units.update(x='m', y='m', z='m')
    with pytest.raises(sc.UnitError):
        canvas._update_camera(limits=_make_limits())


@pytest.mark.parametrize('key', ['position', 'look_at'])
def test_camera_user_vector_can_convert_a_single_field(key):
    vec = sc.vector([0, 1, 2], unit='m')
    canvas = Canvas(camera=Camera(**{key: vec}))
    canvas.units.update(x='m', y='m', z='cm')
    canvas._update_camera(limits=_make_limits())
    _assert_pos((0, 1, 200), canvas, key)


@pytest.mark.parametrize('key', ['position', 'look_at'])
def test_camera_user_tuple_of_variables(key):
    pos = [
        sc.scalar(2.0, unit='m'),
        sc.scalar(-12.0, unit='m'),
        sc.scalar(44.0, unit='m'),
    ]
    canvas = Canvas(camera=Camera(**{key: pos}))
    canvas.units.update(x='m', y='m', z='m')
    canvas._update_camera(limits=_make_limits())
    _assert_pos((2.0, -12.0, 44.0), canvas, key)


@pytest.mark.parametrize('key', ['position', 'look_at'])
def test_camera_user_tuple_of_variables_unit_conversion(key):
    pos = [
        sc.scalar(2.0, unit='m'),
        sc.scalar(-12.0, unit='cm'),
        sc.scalar(44.0, unit='mm'),
    ]
    canvas = Canvas(camera=Camera(**{key: pos}))
    canvas.units.update(x='m', y='m', z='m')
    canvas._update_camera(limits=_make_limits())
    _assert_pos((2.0, -0.12, 0.044), canvas, key)


@pytest.mark.parametrize('key', ['position', 'look_at'])
def test_camera_user_tuple_of_variables_bad_unit_raises(key):
    pos = [
        sc.scalar(2.0, unit='m'),
        sc.scalar(-12.0, unit='m'),
        sc.scalar(44.0, unit='m'),
    ]
    canvas = Canvas(camera=Camera(**{key: pos}))
    canvas.units.update(x='s', y='s', z='s')
    with pytest.raises(sc.UnitError):
        canvas._update_camera(limits=_make_limits())


@pytest.mark.parametrize('key', ['near', 'far'])
def test_camera_user_float(key):
    value = 6.8
    canvas = Canvas(camera=Camera(**{key: value}))
    canvas._update_camera(limits=_make_limits())
    assert getattr(canvas.camera, key) == value


@pytest.mark.parametrize('key', ['near', 'far'])
def test_camera_user_variable(key):
    value = sc.scalar(15.1, unit='m')
    canvas = Canvas(camera=Camera(**{key: value}))
    canvas.units.update(x='m', y='m', z='m')
    canvas._update_camera(limits=_make_limits())
    assert getattr(canvas.camera, key) == value.value


@pytest.mark.parametrize('key', ['near', 'far'])
def test_camera_user_variable_unit_conversion(key):
    value = sc.scalar(33.0, unit='cm')
    canvas = Canvas(camera=Camera(**{key: value}))
    canvas.units.update(x='m', y='m', z='m')
    canvas._update_camera(limits=_make_limits())
    assert getattr(canvas.camera, key) == 0.33


@pytest.mark.parametrize('key', ['near', 'far'])
def test_camera_user_variable_bad_unit_raises(key):
    value = sc.scalar(6.6, unit='s')
    canvas = Canvas(camera=Camera(**{key: value}))
    canvas.units.update(x='m', y='m', z='m')
    with pytest.raises(sc.UnitError):
        canvas._update_camera(limits=_make_limits())


@pytest.mark.parametrize('key', ['near', 'far'])
def test_camera_user_variable_raises_when_axes_units_are_different(key):
    value = sc.scalar(5.66, unit='m')
    canvas = Canvas(camera=Camera(**{key: value}))
    canvas.units.update(x='m', y='m', z='s')
    with pytest.raises(sc.UnitError, match='All axes must have the same unit'):
        canvas._update_camera(limits=_make_limits())


def test_toggle_axes_3d():
    canvas = Canvas()
    assert canvas.axes_3d.visible
    canvas.toggle_axes3d()
    assert not canvas.axes_3d.visible
    canvas.toggle_axes3d()
    assert canvas.axes_3d.visible


def test_toggle_outline():
    canvas = Canvas()
    canvas.make_outline(limits=_make_limits())
    assert canvas.outline.visible
    canvas.toggle_outline()
    assert not canvas.outline.visible
    canvas.toggle_outline()
    assert canvas.outline.visible


def test_creation_with_title():
    text = 'My title'
    canvas = Canvas(title=text)
    assert canvas.title == text
    html = '<i>My</i> <b>title</b>'
    canvas = Canvas(title=html)
    assert canvas.title == html



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import numpy as np
import pytest
import scipp as sc

from plopp.backends.pythreejs.point_cloud import PointCloud
from plopp.data.testing import scatter


def test_creation():
    da = scatter()
    cloud = PointCloud(data=da, x='x', y='y', z='z')
    assert sc.identical(cloud._data, da)
    assert np.allclose(
        cloud.geometry.attributes['position'].array, da.coords['position'].values
    )


def test_update():
    da = scatter()
    cloud = PointCloud(data=da, x='x', y='y', z='z')
    cloud.update(da)
    assert sc.identical(cloud._data, da)
    cloud.update(da * 2.5)
    assert sc.identical(cloud._data, da * 2.5)


def test_get_limits():
    da = scatter()
    pix = 0.5
    cloud = PointCloud(data=da, x='x', y='y', z='z', pixel_size=pix)
    xlims, ylims, zlims = cloud.get_limits()
    assert sc.identical(xlims[0], da.coords['x'].min() - sc.scalar(0.5 * pix, unit='m'))
    assert sc.identical(xlims[1], da.coords['x'].max() + sc.scalar(0.5 * pix, unit='m'))
    assert sc.identical(ylims[0], da.coords['y'].min() - sc.scalar(0.5 * pix, unit='m'))
    assert sc.identical(ylims[1], da.coords['y'].max() + sc.scalar(0.5 * pix, unit='m'))
    assert sc.identical(zlims[0], da.coords['z'].min() - sc.scalar(0.5 * pix, unit='m'))
    assert sc.identical(zlims[1], da.coords['z'].max() + sc.scalar(0.5 * pix, unit='m'))


def test_get_limits_flat_panel():
    da = scatter()
    da.coords['z'] *= 0.0
    pix = 0.5
    cloud = PointCloud(data=da, x='x', y='y', z='z', pixel_size=pix)
    xlims, ylims, zlims = cloud.get_limits()
    assert sc.identical(xlims[0], da.coords['x'].min() - sc.scalar(0.5 * pix, unit='m'))
    assert sc.identical(xlims[1], da.coords['x'].max() + sc.scalar(0.5 * pix, unit='m'))
    assert sc.identical(ylims[0], da.coords['y'].min() - sc.scalar(0.5 * pix, unit='m'))
    assert sc.identical(ylims[1], da.coords['y'].max() + sc.scalar(0.5 * pix, unit='m'))
    assert sc.identical(zlims[0], sc.scalar(-0.5 * pix, unit='m'))
    assert sc.identical(zlims[1], sc.scalar(0.5 * pix, unit='m'))


def test_pixel_size():
    """
    We make a reference points cloud because additional factors are potentially added to
    the size, depending on the device pixel ratio. Making a reference with a default
    size of 1 makes it easier to test.
    """
    da = scatter()
    reference = PointCloud(data=da, x='x', y='y', z='z', pixel_size=1)
    cloud = PointCloud(data=da, x='x', y='y', z='z', pixel_size=sc.scalar(2, unit='m'))
    assert cloud.material.size == 2.0 * reference.material.size


def test_pixel_size_unit_conversion():
    da = scatter()
    reference = PointCloud(data=da, x='x', y='y', z='z', pixel_size=1)
    cloud = PointCloud(
        data=da, x='x', y='y', z='z', pixel_size=sc.scalar(350, unit='cm')
    )
    assert cloud.material.size == 3.5 * reference.material.size
    with pytest.raises(sc.UnitError):
        PointCloud(data=da, x='x', y='y', z='z', pixel_size=sc.scalar(350, unit='s'))


def test_pixel_size_cannot_have_units_when_spatial_dimensions_have_different_units():
    da = scatter()
    new_x = da.coords['x'].copy()
    new_x.unit = 's'
    da.coords['x'] = new_x
    reference = PointCloud(data=da, x='x', y='y', z='z', pixel_size=1)
    with pytest.raises(ValueError, match='The supplied pixel_size has unit'):
        PointCloud(data=da, x='x', y='y', z='z', pixel_size=sc.scalar(2.5, unit='m'))
    # Ok if no unit supplied
    cloud = PointCloud(data=da, x='x', y='y', z='z', pixel_size=2.5)
    assert cloud.material.size == 2.5 * reference.material.size


def test_creation_raises_when_data_is_not_1d():
    da = scatter()
    da2d = sc.broadcast(da, sizes={**da.sizes, **{'time': 10}})
    with pytest.raises(
        ValueError, match='PointCloud only accepts one dimensional data'
    ):
        PointCloud(data=da2d, x='x', y='y', z='z')


def test_update_raises_when_data_is_not_1d():
    da = scatter()
    cloud = PointCloud(data=da, x='x', y='y', z='z')
    da2d = sc.broadcast(da, sizes={**da.sizes, **{'time': 10}})
    with pytest.raises(
        ValueError, match='PointCloud only accepts one dimensional data'
    ):
        cloud.update(da2d)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import warnings
from functools import partial, reduce
from itertools import groupby
from typing import Literal

from scipp.typing import VariableLike

from ..core import widget_node
from ..core.typing import FigureLike, PlottableMulti
from ..graphics import imagefigure, linefigure
from .common import (
    input_to_nodes,
    preprocess,
    raise_multiple_inputs_for_2d_plot_error,
    require_interactive_backend,
)


class Slicer:
    """
    Class that slices out dimensions from the data and displays the resulting data as
    either a 1D line or a 2D image.

    Note:

    This class primarily exists to facilitate unit testing. When running unit tests, we
    are not in a Jupyter notebook, and the generated figures are not widgets that can
    be placed in the `Box` widget container at the end of the `slicer` function.
    We therefore place most of the code for creating a Slicer in this class, which is
    under unit test coverage. The thin `slicer` wrapper is not covered by unit tests.

    Parameters
    ----------
    obj:
        The input data.
    keep:
        The dimensions to be kept, all remaining dimensions will be sliced. This should
        be a list of dims. If no dims are provided, the last dim will be kept in the
        case of a 2-dimensional input, while the last two dims will be kept in the case
        of higher dimensional inputs.
    autoscale:
        The behavior of the y-axis (1d plots) and color range (2d plots) limits.
        If ``auto``, the limits automatically adjusts every time the data changes.
        If ``grow``, the limits are allowed to grow with time but they do not shrink.
        If ``fixed``, the limits are fixed to the full data range and do not change
        with time.
    coords:
        If supplied, use these coords instead of the input's dimension coordinates.
    vmin:
        The minimum value of the y-axis (1d plots) or color range (2d plots).
    vmax:
        The maximum value of the y-axis (1d plots) or color range (2d plots).
    **kwargs:
        The additional arguments are forwarded to the underlying 1D or 2D figures.
    """

    def __init__(
        self,
        obj: PlottableMulti,
        *,
        keep: list[str] | None = None,
        autoscale: Literal['auto', 'grow', 'fixed'] = 'auto',
        coords: list[str] | None = None,
        vmin: VariableLike | float = None,
        vmax: VariableLike | float = None,
        **kwargs,
    ):
        nodes = input_to_nodes(
            obj, processor=partial(preprocess, ignore_size=True, coords=coords)
        )

        dims = nodes[0]().dims
        if keep is None:
            keep = dims[-(2 if len(dims) > 2 else 1) :]

        if isinstance(keep, str):
            keep = [keep]

        # Ensure all dims in keep have the same size
        sizes = [
            {dim: shape for dim, shape in node().sizes.items() if dim not in keep}
            for node in nodes
        ]
        g = groupby(sizes)
        if not (next(g, True) and not next(g, False)):
            raise ValueError(
                'Slicer plot: all inputs must have the same sizes, but '
                f'the following sizes were found: {sizes}'
            )

        if len(keep) == 0:
            raise ValueError(
                'Slicer plot: the list of dims to be kept cannot be empty.'
            )
        if not all(dim in dims for dim in keep):
            raise ValueError(
                f"Slicer plot: one or more of the requested dims to be kept {keep} "
                f"were not found in the input's dimensions {dims}."
            )

        if autoscale == 'fixed':
            if None not in (vmin, vmax):
                warnings.warn(
                    'Slicer plot: autoscale is set to "fixed", but vmin and vmax '
                    'are also specified. They will override the autoscale setting.',
                    RuntimeWarning,
                    stacklevel=2,
                )
            if vmin is None:
                vmin = reduce(min, [node().data.min() for node in nodes])
            if vmax is None:
                vmax = reduce(max, [node().data.max() for node in nodes])
            autoscale = 'auto'  # Change back to something the figure understands

        from ..widgets import SliceWidget, slice_dims

        self.slider = SliceWidget(
            nodes[0](), dims=[dim for dim in dims if dim not in keep]
        )
        self.slider_node = widget_node(self.slider)
        self.slice_nodes = [slice_dims(node, self.slider_node) for node in nodes]

        ndims = len(keep)
        if ndims == 1:
            make_figure = linefigure
        elif ndims == 2:
            if len(self.slice_nodes) > 1:
                raise_multiple_inputs_for_2d_plot_error(origin='slicer')
            make_figure = imagefigure
        else:
            raise ValueError(
                f'Slicer plot: the number of dims to be kept must be 1 or 2, '
                f'but {ndims} were requested.'
            )

        self.figure = make_figure(
            *self.slice_nodes, autoscale=autoscale, vmin=vmin, vmax=vmax, **kwargs
        )
        self.figure.bottom_bar.add(self.slider)


def slicer(
    obj: PlottableMulti,
    *,
    keep: list[str] | None = None,
    autoscale: Literal['auto', 'grow', 'fixed'] = 'auto',
    coords: list[str] | None = None,
    vmin: VariableLike | float = None,
    vmax: VariableLike | float = None,
    **kwargs,
) -> FigureLike:
    """
    Plot a multi-dimensional object by slicing one or more of the dimensions.
    This will produce one slider per sliced dimension, below the figure.

    Parameters
    ----------
    obj:
        The object to be plotted.
    keep:
        The dimensions to be kept, all remaining dimensions will be sliced. This should
        be a list of dims. If no dims are provided, the last dim will be kept in the
        case of a 2-dimensional input, while the last two dims will be kept in the case
        of higher dimensional inputs.
    autoscale:
        The behavior of the y-axis (1d plots) and color range (2d plots) limits.
        If ``auto``, the limits automatically adjusts every time the data changes.
        If ``grow``, the limits are allowed to grow with time but they do not shrink.
        If ``fixed``, the limits are fixed to the full data range and do not change
        with time.
    coords:
        If supplied, use these coords instead of the input's dimension coordinates.
    vmin:
        The minimum value of the y-axis (1d plots) or color range (2d plots).
    vmax:
        The maximum value of the y-axis (1d plots) or color range (2d plots).
    **kwargs:
        See :py:func:`plopp.plot` for the full list of figure customization arguments.
    """
    require_interactive_backend('slicer')
    return Slicer(
        obj,
        keep=keep,
        autoscale=autoscale,
        vmin=vmin,
        vmax=vmax,
        coords=coords,
        **kwargs,
    ).figure



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)


class PlotlyBackend:
    def is_interactive(self):
        """
        Returns ``True`` if the backend currently in use allows for interactive figures.
        """
        return True

    def canvas2d(self, *args, **kwargs):
        """
        See :class:`canvas.Canvas` for details.
        """
        from .canvas import Canvas as CanvasPlotly

        return CanvasPlotly(*args, **kwargs)

    def line(self, *args, **kwargs):
        """
        See :class:`line.Line` for details.
        """
        from .line import Line as LinePlotly

        return LinePlotly(*args, **kwargs)

    def figure1d(self, *args, **kwargs):
        """
        See :class:`figure.Figure` for details.
        """
        from .figure import Figure as FigPlotly

        return FigPlotly(*args, **kwargs)


__all__ = ["PlotlyBackend"]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from .basefig import BaseFig
from .camera import Camera
from .colormapper import ColorMapper
from .figure import figure1d, figure2d, figure3d
from .imageview import ImageView, imagefigure
from .lineview import LineView, linefigure
from .scatter3dview import Scatter3dView, scatter3dfigure
from .scatterview import ScatterView, scatterfigure
from .tiled import tiled

__all__ = [
    'BaseFig',
    'Camera',
    'ColorMapper',
    'ImageView',
    'Scatter3dView',
    'ScatterView',
    'figure1d',
    'figure2d',
    'figure3d',
    'imagefigure',
    'LineView',
    'linefigure',
    'scatter3dfigure',
    'scatterfigure',
    'tiled',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

import numpy as np
import scipp as sc

from ..core.limits import find_limits, fix_empty_range
from ..core.utils import merge_masks


def _none_min(*args: float) -> float:
    return min(x for x in args if x is not None)


def _none_max(*args: float) -> float:
    return max(x for x in args if x is not None)


@dataclass
class BoundingBox:
    """
    A bounding box in 2D space.
    """

    xmin: float | None = None
    xmax: float | None = None
    ymin: float | None = None
    ymax: float | None = None

    def union(self, other: BoundingBox) -> BoundingBox:
        """
        Return the union of this bounding box with another one.
        """

        return BoundingBox(
            xmin=_none_min(self.xmin, other.xmin),
            xmax=_none_max(self.xmax, other.xmax),
            ymin=_none_min(self.ymin, other.ymin),
            ymax=_none_max(self.ymax, other.ymax),
        )


def axis_bounds(
    keys: tuple[str, str],
    x: sc.DataArray,
    scale: Literal['linear', 'log'],
    pad=False,
) -> dict[str, float]:
    """
    Find sensible limits for an axis, depending on linear or log scale.

    Parameters
    ----------
    keys:
        The keys to use for constructing a bounding box. The keys should be
        ``('xmin', 'xmax')`` for the horizontal axis, and ``('ymin', 'ymax')`` for the
        vertical axis.
    x:
        The data array to find limits for.
    scale:
        The scale of the axis (linear or log).
    pad:
        Whether to pad the limits.
    """
    values = fix_empty_range(find_limits(x, scale=scale, pad=pad))
    bounds = dict(zip(keys, (val.value for val in values), strict=True))
    return bounds


def make_line_data(data: sc.DataArray, dim: str) -> dict:
    """
    Prepare data for plotting a line.
    This includes extracting the x and y values, and optionally the error bars and masks
    from the data array.
    This also handles the case where the data array is a histogram, in which case the
    first bin edge is repeated.

    Parameters
    ----------
    data:
        The data array to extract values from.
    dim:
        The dimension along which to extract values.
    """
    x = data.coords[dim]
    y = data.data
    hist = len(x) != len(y)
    error = None
    xvalues = np.asarray(x.values)
    yvalues = np.asarray(y.values)
    values = {'x': xvalues, 'y': yvalues}
    mask = {'x': xvalues, 'y': np.full(y.shape, np.nan), 'visible': False}
    if data.variances is not None:
        error = {
            'x': np.asarray(sc.midpoints(x).values) if hist else xvalues,
            'y': yvalues,
            'e': np.asarray(sc.stddevs(y).values),
        }
    if len(data.masks):
        one_mask = np.asarray(merge_masks(data.masks).values)
        mask = {
            'x': xvalues,
            'y': np.where(one_mask, yvalues, np.nan),
            'visible': True,
        }
    if hist:
        for array in (values, mask):
            array['y'] = np.concatenate([array['y'][0:1], array['y']])
    return {'values': values, 'stddevs': error, 'mask': mask, 'hist': hist}



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from collections.abc import Callable
from typing import Any

from .node import Node


def node(func: Callable) -> Callable:
    """
    Create a :class:`Node` from a callable.
    This can also be used as a decorator.

    Parameters
    ----------
    func:
        The callable to create the :class:`Node`.
    """

    def make_node(*args: Any, **kwargs: Any) -> None:
        return Node(func, *args, **kwargs)

    return make_node


def widget_node(widget) -> Node:
    """
    Create a node from a widget. When data is requested from it, it will return the
    ``.value`` attribute of the supplied widget.
    In addition, all of its children will be notified when the ``value`` of the widget
    changes.

    Parameters
    ----------
    widget:
        The widget used to construct the node. This can be a widget from the
        ``ipywidgets`` library, or a custom widget.
    """
    n = Node(func=lambda: widget.value)
    n.name = f'Widget <{type(widget).__name__}: {type(widget.value).__name__}>'
    # TODO: Our custom widgets have a '_plopp_observe' method instead of 'observe'
    # because inheriting from VBox causes errors when overriding the 'observe' method
    # (see https://bit.ly/3SggPVS).
    observe_func = (
        widget._plopp_observe_ if hasattr(widget, '_plopp_observe_') else widget.observe
    )
    observe_func(n.notify_children, names="value")
    return n



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

import uuid
from abc import abstractmethod
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .node import Node


class View:
    """
    A (typically graphical) representation of the data.
    A view must be attached to one or more :class:`Node`.
    Upon receiving a notification from a parent node, it usually requests data from that
    node in order to display it visually.

    Parameters
    ----------
    *nodes:
        The nodes that are attached to the view.
    """

    def __init__(self, *nodes: Node) -> None:
        self._id = uuid.uuid4().hex
        self.graph_nodes = {}
        for node in nodes:
            node.add_view(self)
        self.artists = {}

    @property
    def id(self) -> str:
        """
        The unique id of the view.
        """
        return self._id

    def notify_view(self, message: dict[str, Any]) -> None:
        """
        When a notification is received, request data from the corresponding parent node
        and update the relevant artist.

        Parameters
        ----------
        *message:
            The notification message containing the node id it originated from.
        """
        node_id = message["node_id"]
        new_values = self.graph_nodes[node_id].request_data()
        self.update(**{node_id: new_values})

    @abstractmethod
    def update(self, *args: Any, **kwargs: Any) -> None:
        """
        Update function which is called when a notification is received.
        This has to be overridden by any child class.
        """
        ...

    def render(self) -> None:
        """
        At the end of figure creation, this function is called to request data from
        all parent nodes and draw the figure.
        """
        self.update(
            **{node.id: node.request_data() for node in self.graph_nodes.values()}
        )
